# -*- coding: utf-8 -*-

import math

import matplotlib.pyplot as plt
from matplotlib.pyplot import get_cmap
from matplotlib.colors import rgb2hex
from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D
import numpy as np
import pandas as pd
from pandas.plotting import parallel_coordinates
import scipy.cluster.hierarchy as shc
from scipy.stats import f_oneway


def plot_clustering_results(lst, title, metric, output, errorbar=None,
                            annotation=None):
    """
    Function to plot goodness of fit indicators resulting from a clustering
    model. Resulting plot will be saved in the output folder specified in
    function's arguments.

    Args:
        lst (List):                 List of values to plot.
        title (str):                Title of the plot.
        metric (str):               Metric name.
        output (str):               Output filename.
        errorbar (List):            List of values to plot as errorbar
                                    (CI, SD, etc.).
        annotation (str, optional): Annotation to add directly on the plot.
                                    Defaults to None.
    """

    # Plotting data.
    fig = plt.figure(figsize=(10, 7))
    axes = fig.add_subplot(111)

    with plt.rc_context(
        {"font.size": 10, "font.weight": "bold", "axes.titleweight": "bold"}
    ):
        axes.plot(range(2, len(lst) + 2), lst)

        # Add error bar if provided.
        if errorbar is not None:
            assert len(errorbar) == len(
                lst
            ), "Values to plot and error bars are not of the same length "
            "[{} and {}]".format(len(lst), len(errorbar))

            axes.errorbar(
                range(2, len(errorbar) + 2),
                lst,
                yerr=errorbar,
                ecolor="black",
                elinewidth=1,
                fmt="o",
                color="black",
                barsabove=True,
            )

        # Set parameters.
        plt.xticks(range(2, len(lst) + 2))
        plt.title(f"{title}")
        plt.xlabel("Number of clusters")
        plt.ylabel(f"{metric}")
        axes.spines[["top", "right"]].set_visible(False)
        axes.spines["bottom"].set(linewidth=1.5)
        axes.spines["left"].set(linewidth=1.5)

        # Annotating
        if annotation is not None:
            if axes.get_ylim()[0] < 0:
                y_pos = axes.get_ylim()[0] * 0.95
            else:
                y_pos = axes.get_ylim()[1] * 0.95

            plt.text(x=axes.get_xlim()[1] * 0.5, y=y_pos, s=f"{annotation}")

        plt.savefig(f"{output}")
        plt.close()


def plot_dendrogram(X, output, title="Dendrograms", annotation=None):
    """
    Function to plot a dendrogram plot showing hierarchical clustering. Useful
    to visually determine the appropriate number of clusters.
    Adapted from:
    https://towardsdatascience.com/cheat-sheet-to-implementing-7-methods-for-selecting-optimal-number-of-clusters-in-python-898241e1d6ad

    Args:
        X (DataFrame):                  Data on which clustering will be
                                        performed.
        output (str):                   Output filename and path.
        title (str, optional):          Title for the plot. Defaults to
                                        'Dendrograms'.
        annotation (str, optional):     Annotation to add directly on the plot.
                                        Defaults to None.
    """

    fig = plt.figure(figsize=(10, 7))
    axes = fig.add_subplot(111)

    with plt.rc_context(
        {"font.size": 10, "font.weight": "bold", "axes.titleweight": "bold"}
    ):
        axes.set_title(f"{title}")

        shc.dendrogram(shc.linkage(X, method="ward"))

        if annotation is not None:
            plt.text(
                x=fig.axes[0].get_xlim()[1] * 0.10,
                y=fig.axes[0].get_ylim()[1] * 0.85,
                s=f"{annotation}",
            )

        axes.set_xticks([])
        axes.set_yticks([])
        axes.spines[["top", "right", "left"]].set_visible(False)
        axes.spines["bottom"].set(linewidth=1.5)

        plt.savefig(f"{output}")
        plt.close()


def sort_int_labels_legend(ax, title=None):
    """
    Function automatically reorder numerically labels with matching handles in
    matplotlib legend.

    Args:
        ax:                     Matplotlib Axes.
        title (str, optional):  Title of the legend.

    Returns:
        ax.legend:              Axes legend object.
    """

    # Fetching handles and tags from matplotlib axes.
    handles, tags = ax.get_legend_handles_labels()

    # Converting tags to int if they are not already.
    if type(tags[0]) is str:
        tags = [int(float(x)) for x in tags]

    # Sorting tags and make handles follow the same ordering.
    tag, handle = zip(*sorted(zip(tags, handles)))

    if title is not None:
        return ax.legend(handle, tag, title=title)

    else:
        return ax.legend(handle, tag)


def plot_parallel_plot(
    X, labels, output, mean_values=False, cmap='magma',
    title="Parallel Coordinates plot."
):
    """
    Function to plot a parallel coordinates plot to visualize differences
    between clusters. Useful to highlight significant changes between clusters
    and interpret them.
    Adapted from:
    https://towardsdatascience.com/the-art-of-effective-visualization-of-multi-dimensional-data-6c7202990c57

    Args:
        X (pd.DataFrame):                   Input dataset of shape (S, F).
        labels (Array):                     Array of hard membership value.
                                            (S, ).
        output (str):                       Filename of the png file.
        cmap (str, optional):               Colormap to use for the plot.
                                            Defaults to 'magma'. See
                                            https://matplotlib.org/stable/tutorials/colors/colormaps.html
        title (str, optional):              Title of the plot.
                                            Defaults to 'Parallel Coordinates
                                            plot.'.
    """

    labels = labels + 1

    columns = list(X.columns)
    columns.append("Clusters")

    df = pd.concat([X, pd.DataFrame(labels, columns=["Clusters"])], axis=1)

    if mean_values:
        # Calculating mean values for each features for each clusters.
        final_df = pd.DataFrame()
        i = 0
        for col in X.columns:
            mean = list()
            for k in np.unique(labels):
                mean.append(df.loc[df["Clusters"] == k, col].mean())
            final_df.insert(i, col, mean)
            i += 1
        final_df.insert(i, "Clusters", np.unique(labels))
    else:
        indexes = np.random.choice(df.shape[0], size=500, replace=False)
        final_df = df.iloc[indexes]

    # Setting color palette.
    cmap = get_cmap(cmap, len(np.unique(labels)))
    colors = [rgb2hex(cmap(i)) for i in range(cmap.N)]

    fig = plt.figure(figsize=(15, 7))
    ax = fig.add_subplot(111)

    with plt.rc_context(
        {"font.size": 10, "font.weight": "bold", "axes.titleweight": "bold"}
    ):
        parallel_coordinates(final_df, "Clusters", ax=ax, color=colors)
        sort_int_labels_legend(ax, title="Cluster #")
        ax.set_title(f"{title}")
        ax.grid(False)
        ax.spines[["left", "right", "top", "bottom"]].set(linewidth=1.5)

        ax.figure.autofmt_xdate()

        plt.savefig(f"{output}")
        plt.close()


def radar_plot(X, labels, output, frame='circle', title="Radar plot",
               cmap='magma'):
    """
    Function to plot a radar plot for all features in the original dataset
    stratified by clusters. T-test between clusters' mean within a feature is
    also computed and annotated directly on the plot. When plotting a high
    number of clusters, plotting of significant annotation is polluting the
    plot, will be fixed in the future.

    Args:
        X (DataFrame):                      Input dataset of shape (S, F).
        labels (Array):                     Array of hard membership value
                                            (S, ).
        output (str):                       Filename of the png file.
        frame (str, optional):              Shape of the radar plot. Defaults
                                            to 'circle'. Choices are 'circle'
                                            or 'polygon'.
        title (str, optional):              Title of the plot.
                                            Defaults to 'Radar plot'.
        cmap (str, optional):               Colormap to use for the plot.
                                            Defaults to 'magma'. See
                                            https://matplotlib.org/stable/tutorials/colors/colormaps.html
    """
    # Setting color palette.
    cmap = get_cmap(cmap, len(np.unique(labels)))
    colors = [rgb2hex(cmap(i)) for i in range(cmap.N)]

    # Make labels start at 1 rather than 0, better for viz.
    labels = labels + 1

    # Axis labels.
    var_labels = X.columns.tolist()
    var_labels.append(var_labels[0])

    # Computing ANOVA for each features.
    anova = []
    i = 0
    for col in X.columns:
        f, p = f_oneway(*[X.loc[labels == k, col] for k in np.unique(labels)])
        anova.append(p)
        i += 1

    # Computing mean values for each features for each clusters.
    mean_df = pd.DataFrame()
    i = 0
    for col in X.columns:
        mean = list()
        for k in np.unique(labels):
            mean.append(X.loc[labels == k, col].mean())
        mean_df.insert(i, col, mean)
        i += 1

    # Computing stds for each features for each clusters.
    std_df = pd.DataFrame()
    i = 0
    for col in X.columns:
        std = list()
        for k in np.unique(labels):
            std.append(X.loc[labels == k, col].std())
        std_df.insert(i, col, std)
        i += 1
    max_val = math.ceil((mean_df + std_df).max().max())
    min_val = math.floor((mean_df - std_df).min().min())

    mean_df.insert(i, "Clusters", np.unique(labels))
    std_df.insert(i, "Clusters", np.unique(labels))

    with plt.rc_context(
        {"font.size": 12, "font.weight": "bold", "axes.titleweight": "bold",
         "font.family": "Sans Serif"}
    ):
        fig = plt.figure(figsize=(12, 12))
        ax = fig.add_subplot(111, polar=True)

        # Set radar plot parameters.
        theta = create_radar_plot(len(X.columns), frame=frame)

        for idx, cluster in enumerate(np.unique(labels)):
            values = mean_df.iloc[idx].drop('Clusters').values.tolist()
            values.append(values[0])
            stds = std_df.iloc[idx].drop('Clusters').values.tolist()
            stds.append(stds[0])
            stds_pos = [np.sum(x) for x in zip(values, stds)]
            stds_neg = [s - d for s, d in zip(values, stds)]
            ax.plot(theta, values, c=colors[idx], linewidth=2,
                    label=f'Cluster {cluster}', markersize=4, zorder=3)
            plot = ax.errorbar(theta, values, yerr=stds, fmt='o-',
                               color=colors[idx], linewidth=0,
                               label=f'Cluster {cluster}')
            ax.fill_between(theta, values, stds_pos, alpha=0.2,
                            color=colors[idx], edgecolor='none',
                            label='_nolegend_')
            ax.fill_between(theta, values, stds_neg, alpha=0.2,
                            color=colors[idx], edgecolor='none',
                            label='_nolegend_')

            plot[-1][0].set_color(colors[idx])

    # Add p-values to the plot.
    for i, p in enumerate(anova):
        if 0.01 < p < 0.05:
            ax.text(theta[i], max_val * 0.95, '*', fontsize=18, color='black',
                    weight='bold', rotation=math.degrees(theta[i]) + 90)
        elif 0.001 < p < 0.01:
            ax.text(theta[i], max_val * 0.95, '**', fontsize=18, color='black',
                    weight='bold', rotation=math.degrees(theta[i]) + 90)
        elif p < 0.001:
            ax.text(theta[i], max_val * 0.95, '***', fontsize=18,
                    color='black', weight='bold',
                    rotation=math.degrees(theta[i]) + 90,
                    ha='center', va='center')

    # Set legend and variables parameters.
    legend = ax.legend(np.unique(labels), loc=(0.95, 0.9), title='Profile #',
                       fontsize=14)
    frame = legend.get_frame()
    frame.set_facecolor('#eef0eb')
    frame.set_edgecolor('gray')
    ax.set_thetagrids(theta * 180 / np.pi, var_labels, zorder=1)
    ax.set_rlabel_position(-36)
    ax.set_ylim(min_val, max_val)
    yticks = np.arange(min_val, max_val, 0.5)
    ax.set_yticks(yticks)

    # Set spines and title parameters.
    for spine in ax.spines.values():
        spine.set_color('black')
    ax.grid(axis='y', color='white', linewidth=1, zorder=3)
    ax.grid(axis='x', color='white', linewidth=.5, zorder=2)
    ax.set_facecolor('#eef0eb')
    ax.set_xticklabels(var_labels)

    ax.set_title(f"{title}", weight='bold', size=16,
                 horizontalalignment='center')

    # Set the position for the labels.
    for label, angle in zip(ax.get_xticklabels(), theta):
        if angle == 0:
            label.set_horizontalalignment('left')
        elif angle == np.pi:
            label.set_horizontalalignment('right')
        elif 0 < angle < np.pi / 2 or angle > 3 * np.pi / 2:
            label.set_horizontalalignment('left')
        else:
            label.set_horizontalalignment('right')

    plt.tight_layout()
    plt.savefig(f"{output}", dpi=300)
    plt.close()


def create_radar_plot(nb_vars, frame='circle'):
    """
    Create a radar chart with `nb_vars` axes.

    Args:
        nb_vars (int):          Number of variables to plot.
        frame (str, optional):  Shape of the radar plot. Defaults to 'circle'.
                                Choices are 'circle' or 'polygon'.

    Returns:
        np.array:               Array of evenly spaced axis angles.
    """

    # Compute evenly spaced axis angles.
    theta = np.linspace(0, 2 * np.pi, nb_vars, endpoint=False)
    theta = np.concatenate((theta, [theta[0]]))

    class RadarTransform(PolarAxes.PolarTransform):

        def transform_path_non_affine(self, path):
            if path._interpolation_steps > 1:
                path = path.interpolated(nb_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = 'radar'
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # Rotate plot to place the first axis at the top.
            self.set_theta_zero_location('N')

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default.

            Args:
                closed (bool, optional): _description_. Defaults to True.
            """
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """
            Override plot so that line is closed by default.
            """
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            if frame == 'circle':
                return Circle((0.5, 0.5), 0.5)
            elif frame == 'polygon':
                return RegularPolygon((0.5, 0.5), nb_vars,
                                      radius=0.5, edgecolor='k')
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == 'circle':
                return super()._gen_axes_spines()
            elif frame == 'polygon':

                spine = Spine(axes=self,
                              spine_type='circle',
                              path=Path.unit_regular_polygon(nb_vars))
                spine.set_transform(
                    Affine2D().scale(0.5).translate(0.5, 0.5) + self.transAxes)
                return {'polar': spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)

    return theta



import inspect
import os
import shutil
import sys

from detect_delimiter import detect
import pandas as pd


"""
Some function comes from the scilpy toolbox. Please see :
https://github.com/scilus/scilpy
"""


def load_df_in_any_format(file):
    """
    Load dataset in any .csv or .xlsx format.
    :param file:    Input file to load.
    :return:
    Pandas dataframe.
    """
    _, ext = os.path.splitext(file)
    if ext == ".csv":
        df = pd.read_csv(file)
    if ext == ".xlsx":
        df = pd.read_excel(file)
    if ext == ".txt":
        with open(file, "r") as f:
            f = f.read()
            delimiter = detect(f, whitelist=["\t", ":", ";", " ", ","])
        df = pd.read_csv(file, sep=delimiter)

    return df


def add_overwrite_arg(p):
    p.add_argument(
        "-f",
        dest="overwrite",
        action="store_true",
        help="Force overwriting of existing output files.",
    )


def add_verbose_arg(p):
    p.add_argument(
        "-v",
        dest="verbose",
        action="store_true",
        help="If true, produce verbose output.",
    )


def assert_input(required, optional=None):
    """
    Function to validate the existence of an input file.
    From the scilpy toolbox : https://github.com/scilus/scilpy
    :param p:           Parser
    :param required:    Paths to assert the existence
    :param optional:    Paths to assert optional arguments
    :return:
    """

    def check(path):
        if not os.path.isfile(path):
            sys.exit("File {} does not exist.".format(path))

    if isinstance(required, str):
        required = [required]

    if isinstance(optional, str):
        optional = [optional]

    for file in required:
        check(file)
    for file in optional or []:
        if file is not None:
            check(file)


def assert_output(overwrite, required, optional=None, check_dir=True):
    """
    Validate that output exist and force the use of -f.
    From the scilpy toolbox : https://github.com/scilus/scilpy
    :param overwrite:       Overwrite argument value (true or false).
    :param required:    Required paths to assert.
    :param optional:    Optional paths to assert.
    :param check_dir:   Validate if output directory exist.
    :return:
    """

    def check(path):
        if os.path.isfile(path) and not overwrite:
            sys.exit(
                "Output file {} exists. Select the -f to force "
                "overwriting of the existing file.".format(path)
            )

        if check_dir:
            path_dir = os.path.dirname(path)
            if path_dir and not os.path.isdir(path_dir):
                sys.exit(
                    "Directory {} is not created for the output file."
                    .format(path_dir)
                )

    if isinstance(required, str):
        required = [required]
    if isinstance(optional, str):
        optional = [optional]

    for file in required:
        check(file)
    for file in optional or []:
        if file:
            check(file)


def assert_output_dir_exist(overwrite, required, optional=None,
                            create_dir=True):
    """
    Validate the existence of the output directory.
    From the scilpy toolbox : https://github.com/scilus/scilpy
    :param overwrite:       Overwrite argument value (true or false).
    :param required:        Required paths to validate.
    :param optional:        Optional paths to validate.
    :param create_dir:      Option to create the directory if it does not
                            already exist.
    :return:
    """

    def check(path):
        if not os.path.isdir(path):
            if not create_dir:
                sys.exit(
                    "Output directory {} does not exist. Use create_dir = "
                    "True.".format(path)
                )
            else:
                os.makedirs(path, exist_ok=True)
        if os.listdir(path):
            if not overwrite:
                sys.exit(
                    "Output directory {} is not empty. Use -f to overwrite the"
                    "existing content.".format(path)
                )
            else:
                for file in os.listdir(path):
                    file_path = os.path.join(path, file)
                    try:
                        if os.path.isfile(file_path):
                            os.unlink(file_path)
                        elif os.path.isdir(file_path):
                            shutil.rmtree(file_path)
                    except Exception as e:
                        print(e)

    if isinstance(required, str):
        required = [required]
    if isinstance(optional, str):
        optional = [optional]

    for cur_dir in required:
        check(cur_dir)
    for opt_dir in optional or []:
        if opt_dir:
            check(opt_dir)


def get_data_dir():
    """
    Return a data directory within the NeuroStatX repository
    :return:
    data_dir: data path
    """
    import neurostatx

    module_path = inspect.getfile(neurostatx)

    data_dir = os.path.join(os.path.dirname(
                            os.path.dirname(module_path)) + "/data/")

    return data_dir



# -*- coding: utf-8 -*-

from enum import Enum
from strenum import StrEnum


class DistanceMetrics(StrEnum, Enum):
    braycurtis = ("braycurtis",)
    canberra = ("canberra",)
    chebyshev = ("chebyshev",)
    cityblock = ("cityblock",)
    correlation = ("correlation",)
    cosine = ("cosine",)
    dice = ("dice",)
    euclidean = ("euclidean",)
    hamming = ("hamming",)
    jaccard = ("jaccard",)
    jensenshannon = ("jensenshannon",)
    kulczynski1 = ("kulczynski1",)
    mahalanobis = ("mahalanobis",)
    matching = ("matching",)
    minkowski = ("minkowski",)
    rogerstanimoto = ("rogerstanimoto",)
    russellrao = ("russellrao",)
    seuclidean = ("seuclidean",)
    sokalmichener = ("sokalmichener",)
    sokalsneath = ("sokalsneath",)
    sqeuclidean = ("sqeuclidean",)
    yule = "yule"



# -*- coding: utf-8 -*-

import math

from matplotlib import pyplot as plt
from matplotlib.colors import rgb2hex
import numpy as np
import skfuzzy as fuzz
from sklearn.utils.parallel import Parallel, delayed

from neurostatx.clustering.metrics import (
    compute_evaluation_metrics,
    compute_sse,
    compute_gap_stats,
)


# Define process_cluster outside fuzzyCmeans
def process_cluster(X, n_cluster, max_clusters, m, error, maxiter, init,
                    metric):
    """
    Core worker function of fuzzyCmeans(). Compute the clustering, metrics,
    and visualization dataset.

    Args:
        X (Numpy array):                Numpy array with data to cluster
                                        (Subject x Features).
        max_cluster (int, optional):    Maximum number of clusters to fit a
                                        model for. Defaults to 10.
        m (float, optional):            Exponentiation value to apply on the
                                        membership function. Defaults to 2.
        error (float, optional):        Stopping criterion. Defaults to 1E-6.
        maxiter (int, optional):        Maximum iteration value.
                                        Defaults to 1000.
        init (2d array, optional):      Initial fuzzy c-partitioned matrix.
                                        Defaults to None.
        metric (str, optional):         Distance metric to use to compute
                                        intra/inter subjects/clusters distance.
                                        Defaults to euclidean.

    Returns:
        n_cluster:                      Number of cluster.
        p:                              Number of iterations run.
        fpc:                            Fuzzy partition coefficient.
        cntr:                           Cluster centroids array.
        u:                              Membership array.
        wss:                            Within-cluster Sum of Square Error.
        ss:                             Silhouette Coefficient Score.
        chi:                            Calinski-Harabasz Index.
        dbi:                            Davies-Bouldin Index.
        gap:                            GAP statistic.
        sk:                             GAP standard error.
        xpts_for_viz:                   X data points for visualization.
        ypts_for_viz:                   Y data points for visualization.
        cluster_membership_for_viz:     Hard membership value for visualization
                                        data points.
    """

    if n_cluster <= max_clusters:
        if init is not None:
            init_mat = init[n_cluster - 2]
        else:
            init_mat = None

        cntr, u, _, _, _, p, fpc = fuzz.cmeans(
            X.T,
            n_cluster,
            m=m,
            error=error,
            maxiter=maxiter,
            metric=metric,
            init=init_mat,
            seed=0)

        cluster_membership = np.argmax(u, axis=0)

        ss, chi, dbi = compute_evaluation_metrics(X, cluster_membership,
                                                  metric=metric)
        labels = np.argmax(u, axis=0)
        wss = compute_sse(X, cntr, labels)

        gap, sk = compute_gap_stats(
            X,
            wss,
            nrefs=100,
            n_cluster=n_cluster,
            m=m,
            error=error,
            metric=metric,
            maxiter=maxiter,
            init=None,
        )

        xpts = X[:, 0]
        ypts = X[:, 1]

        if X.shape[0] > 500:
            indices = np.random.choice(X.shape[0], size=500, replace=False)
            xpts_for_viz = xpts[indices]
            ypts_for_viz = ypts[indices]
            cluster_membership_for_viz = cluster_membership[indices]
        else:
            xpts_for_viz = xpts
            ypts_for_viz = ypts
            cluster_membership_for_viz = cluster_membership

        return (
            n_cluster,
            p,
            fpc,
            cntr,
            u,
            wss,
            ss,
            chi,
            dbi,
            gap,
            sk,
            xpts_for_viz,
            ypts_for_viz,
            cluster_membership_for_viz,
        )

    else:
        return None


def fuzzyCmeans(
    X,
    max_cluster=10,
    m=2,
    error=1e-6,
    maxiter=1000,
    init=None,
    metric="euclidean",
    output="./",
    processes=1,
    verbose=False,
):
    """
    Fuzzy C-Means clustering function. Iteratively test and report statistics
    on multiple number of clusters. Based on documentation found here :
    https://pythonhosted.org/scikit-fuzzy/auto_examples/plot_cmeans.html

    Args:
        X (Numpy array):                Numpy array with data to cluster
                                        (Subject x Features).
        max_cluster (int, optional):    Maximum number of clusters to fit a
                                        model for. Defaults to 10.
        m (float, optional):            Exponentiation value to apply on the
                                        membership function. Defaults to 2.
        error (float, optional):        Stopping criterion. Defaults to 1E-6.
        maxiter (int, optional):        Maximum iteration value. Defaults to
                                        1000.
        init (folder, optional):        Folder containing the c-partitioned
                                        matrices for each cluster number.
                                        Defaults to None.
        metric (str, optional):         Distance metric to use to compute
                                        intra/inter subjects/clusters distance.
                                        Defaults to euclidean.
        output (String, optional):      Output folder. Defaults to './'.
        processes (int, optional):      Number of processes to use for
                                        multiprocessing. Defaults to 1.
        verbose (bool, optional):       Set verbose output. Defaults to False.

    Returns:
        cntr:                           List of cluster centroids arrays for
                                        each model.
        u:                              List of membership arrays for each
                                        model.
        d:                              List of distance arrays for each model.
        wss:                            List of Within-cluster Sum of Square
                                        Error value for each model.
        fpc:                            List of fuzzy partition coefficient
                                        value for each model.
        ss:                             List of Silhouette Coefficient Score
                                        for each model.
        chi:                            List of Calinski-Harabasz Index for
                                        each model.
        dbi:                            List of Davies-Bouldin Index for each
                                        model.
        gap:                            List of GAP statistic for each model.
    """

    num_clusters = max_cluster
    grid = math.ceil(math.sqrt(num_clusters))

    # Spawing multiprocessing using sklearn parallel and delayed function.
    results = Parallel(n_jobs=processes, verbose=verbose)(
        delayed(process_cluster)(
            X,
            k,
            max_clusters=num_clusters,
            m=m,
            error=error,
            maxiter=maxiter,
            init=init,
            metric=metric
        )
        for k in range(2, num_clusters + 2)
    )

    # Process results.
    valid_results = [r for r in results if r is not None]

    # Prepare outputs.
    _, _, fpc, cntr, u, wss, ss, chi, dbi, gap, sk, _, _, _ = zip(
        *valid_results)

    # Save viz.
    fig, axes = plt.subplots(grid, grid, figsize=(8, 8))

    for result in valid_results:
        (
            n_cluster,
            p,
            fpc_,
            cntr_,
            _,
            _,
            _,
            _,
            _,
            _,
            _,
            xpts_for_viz,
            ypts_for_viz,
            cluster_membership_for_viz,
        ) = result

        ax = axes[(n_cluster - 2) // grid, (n_cluster - 2) % grid]

        cmap = plt.get_cmap("plasma", n_cluster)
        colors = [rgb2hex(cmap(i)) for i in range(cmap.N)]

        for j in range(n_cluster):
            ax.plot(
                xpts_for_viz[cluster_membership_for_viz == j],
                ypts_for_viz[cluster_membership_for_viz == j],
                ".",
                color=colors[j],
            )

        for pt in cntr_:
            ax.plot(pt[0], pt[1], "rs")

        ax.set_title(
            "Clusters = {0}; FPC = {1:.2f}\nIterations = {iteration}"
            .format(
                n_cluster, fpc_, iteration=p
            ),
            fontdict={"fontsize": 8},
        )
        ax.axis("off")

    # Removing unused axis.
    for ax in axes.flat[(num_clusters - 1):]:
        ax.remove()

    plt.tight_layout()
    plt.savefig(f"{output}/viz_multiple_cluster_nb.png")
    plt.close()

    return cntr, u, wss, fpc, ss, chi, dbi, gap, sk



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Import required libraries.
import coloredlogs
import dill as pickle
import logging
import sys

from cyclopts import App, Parameter
from factor_analyzer.factor_analyzer import calculate_bartlett_sphericity
from factor_analyzer.factor_analyzer import calculate_kmo
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.model_selection import train_test_split
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (
    assert_input,
    assert_output_dir_exist,
    load_df_in_any_format,
)
from neurostatx.io.viz import flexible_barplot
from neurostatx.utils.preprocessing import merge_dataframes
from neurostatx.utils.factor import (
    RotationTypes,
    MethodTypes,
    horn_parallel_analysis,
    efa,
)


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def ExploratoryFA(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            group="Essential Files Options",
        ),
    ] = "./ResultsEFA/",
    rotation: Annotated[
        RotationTypes,
        Parameter(
            group="Factorial Analysis parameters",
        ),
    ] = RotationTypes.promax,
    method: Annotated[
        MethodTypes,
        Parameter(
            group="Factorial Analysis parameters",
        ),
    ] = MethodTypes.minres,
    nb_factors: Annotated[
        int,
        Parameter(
            "--nb_factors",
            group="Factorial Analysis parameters",
        ),
    ] = None,
    train_dataset_size: Annotated[
        float,
        Parameter(
            "--train_dataset_size",
            group="Factorial Analysis parameters",
        ),
    ] = 0.5,
    random_state: Annotated[
        int,
        Parameter(
            "--random_state",
            group="Factorial Analysis parameters",
        ),
    ] = 1234,
    mean: Annotated[
        bool,
        Parameter(
            "--mean",
            group="Imputing parameters",
        ),
    ] = False,
    median: Annotated[
        bool,
        Parameter(
            "--median",
            group="Imputing parameters",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """EXPLORATORY FACTORIAL ANALYSIS
    ------------------
    ExploratoryFA is a script that can be used to perform an
    exploratory factorial analysis (EFA).

    In the case of performing only an EFA (use the flag --use_only_efa), the
    script will use Horn's parallel analysis to determine the optimal number
    of factors to extract from the data. Then the final EFA model will be
    fitted using the provided rotation and method.

    It is also possible to perform EFA on a training dataset and export the
    test dataset to be used for further analysis (e.g. ConfirmatoryFA). The
    script will output the EFA model, the loadings, communalities, and the
    transformed dataset.

    INPUT SPECIFICATIONS
    --------------------
    Dataset can contain multiple descriptive rows before the variables of
    interest. Simply specify the number of descriptive rows using
    --desc-columns. Rows with missing values will be removed by
    default, please select the mean or median option to impute missing data
    (be cautious when doing this).

    REFERENCES
    ----------
    [1] [Costa, V., & Sarmento, R. Confirmatory Factor
    Analysis.](https://arxiv.org/ftp/arxiv/papers/1905/1905.05598.pdf)

    [2] [Whether to use EFA or CFA to predict latent variables
    scores.](https://stats.stackexchange.com/questions/346499/whether-to-use-efa-or-cfa-to-predict-latent-variables-scores)

    [3] [Comparison of factor score estimation
    methods](https://github.com/gagnonanthony/NeuroStatX/pull/11)

    EXAMPLE USAGE
    -------------
    ::

        ExploratoryFA --in-dataset df --id-column IDs --out-folder results_FA/
        --rotation promax --method ml --train_dataset_size 0.5 -v -f -s

    Parameters
    ----------
    in_dataset : List[str]
        Input dataset(s) to use in the factorial analysis. If multiple files
        are provided as input, will be merged according to the subject id
        columns. For multiple inputs, use this: --in-dataset df1 --in-dataset
        df2 [...]
    id_column : str
        Name of the column containing the subject's ID tag. Required for proper
        handling of IDs and merging multiple datasets.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset to
        exclude in statistics and descriptive tables.
    out_folder : str, optional
        Path of the folder in which the results will be written. If not
        specified, current folder and default name will be used (e.g. =
        ./output/).
    rotation : RotationTypes, optional
        Select the type of rotation to apply on your data.
    method : MethodTypes, optional
        Select the method for fitting the data.
    nb_factors : int, optional
        Specify the number of factors to extract from the data. If not
        specified, the script will use Horn's parallel analysis to determine
        the optimal number of factors.
    train_dataset_size : float, optional
        Specify the proportion of the input dataset to use as training dataset
        in the EFA. (value from 0 to 1)
    random_state : int, optional
        Random seed for reproducibility.
    mean : bool, optional
        Impute missing values in the original dataset based on the column mean.
    median : bool, optional
        Impute missing values in the original dataset based on the column
        median.
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, save the parameters used in the analysis in a text file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    logging.info(
        "Validating input files and creating output folder {}"
        .format(out_folder)
    )
    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading dataset.
    logging.info("Loading {}".format(in_dataset))
    if len(in_dataset) > 1:
        if id_column is None:
            sys.exit(
                "Column name for index matching is required when inputting"
                " multiple dataframes."
            )
        dict_df = {i: load_df_in_any_format(i) for i in in_dataset}
        df = merge_dataframes(dict_df, id_column)
    else:
        df = load_df_in_any_format(in_dataset[0])
    descriptive_columns = [n for n in range(0, desc_columns)]

    # Imputing missing values (or not).
    # Disabled for now, wait until we have a solution to check only columns
    # that will be used in the model.
    # if mean:
    #    logging.info("Imputing missing values using the mean method.")
    #    for column in df.columns:
    #        df[f"{column}"].fillna(df[f"{column}"].mean(), inplace=True)
    # elif median:
    #    logging.info("Imputing missing values using the median method.")
    #    for column in df.columns:
    #        df[f"{column}"].fillna(df[f"{column}"].median(), inplace=True)
    # else:
    #    logging.info(
    #        "No methods selected for imputing missing values. "
    #        "Removing them."
    #    )
    #    df.dropna(inplace=True)

    if train_dataset_size != 1:
        logging.info("Splitting into train and test datasets. Using training "
                     "dataset for EFA.")
        train, test = train_test_split(df, train_size=train_dataset_size,
                                       random_state=random_state)
        train.reset_index(inplace=True, drop=True)
        train.to_excel(f"{out_folder}/train_dataset.xlsx", header=True,
                       index=False)
        test.reset_index(inplace=True, drop=True)
        test.to_excel(f"{out_folder}/test_dataset.xlsx", header=True,
                      index=False)
    else:
        logging.info("Using the full dataset for EFA.")
        train = df

    desc_col = train[train.columns[descriptive_columns]]
    train.drop(df.columns[descriptive_columns], axis=1, inplace=True)

    # Requirement for factorial analysis.
    chi_square_value, p_value = calculate_bartlett_sphericity(train)
    kmo_all, kmo_model = calculate_kmo(train)
    logging.info(
        "Bartlett's test of sphericity returned a p-value of {} and "
        "Keiser-Meyer-Olkin (KMO)"
        "test returned a value of {}.".format(p_value, kmo_model)
    )

    # Horn's parallel analysis.
    suggfactor, suggcomponent = horn_parallel_analysis(
        train.values, out_folder, rotation=None, method=method
    )
    if nb_factors is None:
        nb_factors = suggfactor

    efa_mod, ev, v, scores, loadings, communalities = efa(
        train, rotation=rotation, method=method, nfactors=nb_factors
    )

    # Plot scree plot to determine the optimal number of factors using
    # the Kaiser's method. (eigenvalues > 1)
    plt.scatter(range(1, train.shape[1] + 1), ev)
    plt.plot(range(1, train.shape[1] + 1), ev)
    sns.set_style("whitegrid")
    plt.title("Scree Plot of the eigenvalues for each factor")
    plt.xlabel("Factors")
    plt.ylabel("Eigenvalues")
    plt.grid()
    plt.savefig(f"{out_folder}/scree_plot.png")
    plt.close()

    columns = [f"Factor {i}" for i in range(1, nb_factors + 1)]

    # Export EFA and CFA transformed data.
    efa_out = pd.DataFrame(
        scores, columns=columns
    )
    efa_out = pd.concat([desc_col, efa_out], axis=1)
    efa_out.to_excel(f"{out_folder}/EFA_scores.xlsx", header=True,
                     index=False)

    # Plot correlation matrix between all raw variables.
    corr = pd.DataFrame(efa_mod.corr_, index=train.columns,
                        columns=train.columns)
    mask = np.triu(np.ones_like(corr, dtype=bool))
    f, ax = plt.subplots(figsize=(11, 9))
    ax = sns.heatmap(
        corr,
        mask=mask,
        cmap="BrBG",
        vmax=1,
        vmin=-1,
        center=0,
        square=True,
        annot=True,
        linewidth=0.5,
        fmt=".1f",
        annot_kws={"size": 8},
    )
    ax.set_title("Correlation Heatmap of raw variables.")
    plt.tight_layout()
    plt.savefig(f"{out_folder}/Heatmap.png")
    plt.close()

    # Plot EFA loadings in a barplot.
    efa_loadings = pd.DataFrame(
        loadings, columns=columns, index=train.columns
    )
    efa_communalities = pd.DataFrame(
        communalities, columns=["Communalities"], index=train.columns
    )

    flexible_barplot(
        efa_loadings,
        nb_factors,
        title="Loadings values for the EFA",
        output=f"{out_folder}/barplot_loadings.png",
        ylabel="Loading value",
    )

    # Export EFA loadings for all variables.
    eigen_table = pd.DataFrame(
        ev,
        index=[f"Factor {i}" for i in range(1, len(train.columns) + 1)],
        columns=["Eigenvalues"],
    )
    eigen_table.to_excel(
        f"{out_folder}/eigenvalues.xlsx", header=True, index=True
    )
    efa_loadings.to_excel(
        f"{out_folder}/loadings.xlsx", header=True, index=True
    )
    efa_communalities.to_excel(
        f"{out_folder}/communalities.xlsx", header=True, index=True
    )

    # Export EFA model.
    with open(f"{out_folder}/EFA_model.pkl", "wb") as f:
        pickle.dump(efa_mod, f)


if __name__ == "__main__":
    app()



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def determine_layout(nb_axes):
    """
    Returns the optimal number of rows and columns for the bar plot.

    Args:
        nb_axes (int):      Number of axes to plot.

    Returns:
        int, int:           Number of rows and columns.
    """

    num_rows = int(np.sqrt(nb_axes))
    num_cols = int(np.ceil(nb_axes / num_rows))

    return num_rows, num_cols


def flexible_barplot(
    df, nb_axes, output, cmap='magma', title='Barplot', xlabel=None,
    ylabel=None
):
    """
    Function to generate a bar plot with multiple axes in a publication-ready
    style.

    Args:
        values (pd.DataFrame):          Dataframe with the values to plot. The
                                        index represents the x-axis and the
                                        columns the y-axis.
        num_axes (int):                 Number of axes to plot.
        output (str):                   Output filename.
        cmap (str, optional):           Name of the colormap to use. Defaults
                                        to "magma". See
                                        https://matplotlib.org/stable/tutorials/colors/colormaps.html
        title (str, optional):          Title of the plot.
        xlabel (str, optional):         Label for the x-axis.
        ylabel (str, optional):         Label for the y-axis.
    """

    # Fetch optimal number of rows and columns.
    num_rows, num_cols = determine_layout(nb_axes)

    plotting_parameters = {
        'palette': cmap,
        'saturation': 1,
        'orient': 'v',
    }

    with plt.rc_context(
        {"font.family": "Sans Serif",
         "font.size": 10, "font.weight": "normal", "axes.titleweight": "bold",
         }
    ):
        # Setting up figure and style.
        fig, axes = plt.subplots(
            nrows=num_rows, ncols=num_cols,
            figsize=(8 * num_cols, 6 * num_rows)
        )

        for i, ax in enumerate(axes.flat):
            if i < nb_axes:
                sns.barplot(
                    data=df, x=df.index, hue=df.index, y=df.columns[i],
                    legend=True,
                    ax=ax, **plotting_parameters
                )
                ax.set_title(df.columns[i])
                ax.set_xlabel(xlabel)
                ax.set_ylabel(ylabel)
                ax.grid(False)
                ax.spines[['top', 'right', 'left', 'bottom']].set(linewidth=2)
                plt.setp(ax.get_xticklabels(), rotation=45, ha='right',
                         va='top', rotation_mode='anchor')

                for bars in ax.containers:
                    ax.bar_label(bars, fmt='{:,.3f}', padding=1)

            else:
                ax.axis('off')

        fig.suptitle(title, fontsize=20, fontweight='bold')

        plt.tight_layout()
        plt.savefig(f"{output}", dpi=300, bbox_inches="tight")
        plt.close()


def generate_coef_plot(df, pval, coefname, varname, output, cmap="magma"):
    """
    Function to generate a bar plot with the coefficients and their
    significance.

    Args:
        df (pd.DataFrame):      Dataframe containing the coefficients and their
                                associated variable names.
        coefname (str):         Name of the column containing the coefficients.
        varname (str):          Name of the column containing the variable
                                names.
        output (str):           Output filename.
        cmap (str, optional):   Name of the colormap to use. Defaults to
                                "magma". See
                                https://matplotlib.org/stable/tutorials/colors/colormaps.html
    """
    coef = df[coefname]
    label = ['*' if p < 0.05 else '' for p in pval]
    y = np.arange(0, len(coef))

    plotting_parameters = {
        'data': df,
        'x': coefname,
        'y': varname,
        'palette': cmap,
        'saturation': 1,
        'orient': 'h'
    }

    with plt.rc_context(
        {"font.family": "Sans Serif",
         "font.size": 18, "font.weight": "normal", "axes.titleweight": "bold",
         }
    ):
        fig, ax = plt.subplots(figsize=(10, 10))
        sns.barplot(ax=ax, **plotting_parameters)

        ax.spines[['left', 'right', 'bottom', 'top']].set(linewidth=2)
        ax.set_ylabel('')
        ax.set_xlabel('ß coefficients', fontdict={'fontweight': 'bold',
                                                  'fontsize': 20})

        for i in range(len(coef)):
            if coef[i] < 0:
                plt.text(coef[i], y[i] + 0.05, label[i], ha='right',
                         va='center_baseline', color='black', fontsize=35,
                         weight='bold')
            else:
                plt.text(coef[i], y[i] + 0.05, label[i], ha='left',
                         va='center_baseline', color='black', fontsize=35,
                         weight='bold')

    plt.tight_layout()
    plt.savefig(f"{output}")
    plt.close()


def flexible_hist(df, output, cmap="magma", title="Histogram",
                  xlabel=None, ylabel=None):
    """Function to generate a single histogram representing the distributions
    of all the columns within the dataset in a publication-ready style.

    Args:
        df (pd.DataFrame):          Dataframe containing the values to plot.
        output (str):               Output filename.
        cmap (str, optional):       Colormap. Defaults to "magma". See
                                    https://matplotlib.org/stable/tutorials/colors/colormaps.html
        title (str, optional):      Title of the plot. Defaults to "Histogram".
    """

    with plt.rc_context(
        {"font.family": "Sans Serif",
         "font.size": 12, "font.weight": "bold", "axes.titleweight": "bold"}
    ):

        # Setting up figure and style.
        fig, ax = plt.subplots(figsize=(12, 10))

        sns.histplot(data=df.melt(), x='value', hue='variable', ax=ax,
                     kde=True, palette=cmap, stat='density')

        ax.spines[['top', 'right', 'left', 'bottom']].set(linewidth=1.5)
        ax.set_title(title, fontsize=16, fontweight='bold')
        ax.set_xlabel(xlabel, fontweight='bold')
        ax.set_ylabel(ylabel, fontweight='bold')
        ax.legend(df.columns, title='Variables', fontsize=12,
                  title_fontsize=14)

        plt.tight_layout()
        plt.savefig(f'{output}')
        plt.close()



import unittest
import pandas as pd
import networkx as nx

from neurostatx.network.utils import (get_nodes_and_edges,
                                      filter_node_centroids,
                                      filter_node_subjects,
                                      extract_subject_percentile,
                                      construct_attributes_dict,
                                      fetch_attributes_df,
                                      fetch_edge_data)


class TestFunctions(unittest.TestCase):

    def setUp(self):
        # Generate sample DataFrame
        data = {
            'Subject': ['A', 'B', 'C'],
            'c1': [0.05, 0.1, 0.7],
            'c2': [0.9, 0.1, 0.2],
            'c3': [0.05, 0.8, 0.1],
        }
        self.df = pd.DataFrame(data)

    def test_get_nodes_and_edges(self):
        result_df, subject_list, center_list = get_nodes_and_edges(self.df)
        # Check if returned DataFrame has the correct columns
        self.assertEqual(list(result_df.columns), ['node1', 'node2',
                                                   'membership'])
        # Check if subject_list and center_list are generated correctly
        self.assertEqual(list(subject_list), ['A', 'B', 'C'])
        self.assertEqual(center_list, ['c1', 'c2', 'c3'])

    def test_filter_node_centroids(self):
        self.assertTrue(filter_node_centroids('c1'))
        self.assertFalse(filter_node_centroids('A'))

    def test_filter_node_subjects(self):
        self.assertFalse(filter_node_subjects('c1'))
        self.assertTrue(filter_node_subjects('A'))

    def test_extract_subject_percentile(self):
        mat = self.df.drop('Subject', axis=1).values.T
        label_dict = extract_subject_percentile(mat, 40)
        self.assertEqual(label_dict['c1'].tolist(), [0, 0, 1])
        self.assertEqual(label_dict['c2'].tolist(), [2, 0, 0])

    def test_construct_attributes_dict(self):
        labels = ['c1']
        attributes_dict = construct_attributes_dict(self.df, labels, 'Subject')
        self.assertEqual(attributes_dict, {'A': {'c1': 0.05}, 'B': {'c1': 0.1},
                                           'C': {'c1': 0.7}})

    def test_fetch_attributes_df(self):
        G = nx.Graph()
        G.add_node('A', c1=0.1, c2=0.2)
        G.add_node('B', c1=0.3, c2=0.4)
        G.add_node('C', c1=0.5, c2=0.6)
        df = fetch_attributes_df(G, attributes=['c1'])
        self.assertEqual(df.columns.tolist(), ['c1'])
        self.assertEqual(df.index.tolist(), ['A', 'B', 'C'])
        self.assertEqual(df['c1'].tolist(), [0.1, 0.3, 0.5])

    def test_fetch_edge_data(self):
        result_df, subject_list, center_list = get_nodes_and_edges(self.df)
        G = nx.Graph()
        G = nx.from_pandas_edgelist(result_df, "node1", "node2",
                                    edge_attr="membership")
        df = fetch_edge_data(G)
        expected = pd.DataFrame([[0.05, 0.9, 0.05],
                                 [0.1, 0.1, 0.8],
                                 [0.7, 0.2, 0.1]],
                                index=['A', 'B', 'C'],
                                columns=['Cluster 1', 'Cluster 2',
                                         'Cluster 3'])
        pd.testing.assert_frame_equal(df, expected)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict


download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run(["ApplyModel", "-h"])

    assert ret.success


def test_execution_apply_model(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_dataset = os.path.join(get_home(), "data/factor_data.xlsx")
    out_folder = os.path.join(get_home(), "data/apply_model/")
    model = os.path.join(get_home(), "data/EFA_model.pkl")

    ret = script_runner.run([
        "ApplyModel",
        "--in-dataset", in_dataset,
        "--out-folder", out_folder,
        "--desc-columns", 1,
        "--id-column", "subjectkey",
        "--model", model,
        "-f", "-v", "-s"]
    )

    assert ret.success



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import coloredlogs

from cyclopts import App, Parameter
import pandas as pd
import networkx as nx
from typing_extensions import Annotated
from typing import List

from neurostatx.io.utils import assert_input, assert_output_dir_exist
from neurostatx.network.metrics import weightedpath, PathLengthsMethods
from neurostatx.network.utils import fetch_attributes_df

# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def AverageWeightedPath(
    in_graph: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    label_name: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = None,
    cohort: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = None,
    iterations: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Computational Options",
        ),
    ] = 1000,
    weight: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Computational Options",
        ),
    ] = "membership",
    method: Annotated[
        PathLengthsMethods,
        Parameter(
            show_choices=True,
            show_default=True,
            group="Computational Options",
        ),
    ] = PathLengthsMethods.Dijkstra,
    distribution: Annotated[
        str,
        Parameter(
            show_choices=False,
            show_default=True,
            group="Computational Options",
        ),
    ] = None,
    processes: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Computational Options",
        ),
    ] = 1,
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """AVERAGE WEIGHTED PATH
    ---------------------
    AverageWeightedPath is a script that computes the average
    weighted shortest path length for a group of nodes. The script will
    compare the result against a null distribution generated by randomly
    selecting the same number of nodes and computing the average weighted
    shortest path length. The script will also compute the p-value for the
    comparison.

    NULL DISTRIBUTION
    -----------------
    The null distribution is generated by randomly selecting the same number
    of nodes as the group of interest and computing the average weighted
    shortest path length. This process is repeated for a number of iterations
    (default: 5000) and the resulting distribution is used to compute the
    p-value. Please note that for large graphs, the null distribution could
    take a long time to compute. It is recommended to use the --distribution
    option to provide a pre-computed null distribution of this script has
    already been run on for the same number of nodes (see below). One could
    also reduce the number of iterations to speed up the process and have a
    rough estimate of the p-value.

    COHORT SELECTION
    ----------------
    In some case, the graph network file can contain data from multiple cohort
    at the same time. It might be useful to single out a specific cohort to
    compute the AWP metric. Using --cohort, the script will fetch data from
    the specified cohort and filter the --label-name. The AWP will be
    computed only on subject with the --label-name for the specified cohort.
    In order for the script to run successfully, your graph network needs to
    contain the cohort attributes. If it does not, please add it using the
    AddNodesAttributes script.

    MULTIPROCESSING
    ---------------
    By default, AverageWeightedPath will use 4 CPUs to
    compute the null distribution. This can be changed by setting the
    --processes option.

    METHOD TO COMPUTE THE WEIGHTED PATH
    -----------------------------------
    AverageWeightedPath can use different methods to compute the
    weighted path. The default method is Dijkstra. The other methods are
    Bellman-Ford and Floyd-Warshall. Please note that the Floyd-Warshall
    method is not recommended for large graphs. For more details, see [1].

    REFERENCE
    ---------
    [1]
    [NetworkX
    Algorithm](https://networkx.org/documentation/stable/reference/algorithms/generated/networkx.algorithms.shortest_paths.generic.average_shortest_path_length.html#networkx.algorithms.shortest_paths.generic.average_shortest_path_length)

    EXAMPLE USAGE
    -------------
    ::

        AverageWeightedPath --in-graph ./graph.gexf --label-name Group1
        --label-name Group2 --out-folder ./results --iterations 5000
        --weight membership --method Dijkstra

    Parameters
    ----------
    in_graph : str
        Input graph file (.gml).
    label_name : List[str]
        Label(s) name(s) to select group(s) of nodes. Can be multiple.
    out_folder : str, optional
        Output folder where files will be exported.
    cohort: int, optional
        Cohort identifier. If your graph contains data from multiple cohort,
        you can specify the cohort you want to compute the AWP.
    iterations : int, optional
        Number of iterations to perform to generate the null distribution.
    weight : str, optional
        Name of the edge attributes to use during the weighted path computation
    method : PathLengthsMethods, optional
        Method to use for the estimation of the weighted path.
    distribution : str, optional
        Pre-computed null distribution, needs to be an .xlsx file containing
        label name as --label-name.
    processes : int, optional
        Number of processes to use to compute the null distribution.
    verbose : bool, optional
        If true, produce verbose output.
    save_parameters : bool, optional
        If true, will save input parameters to .txt file.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    """

    if verbose:
        verbose = True
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    logging.info("Loading graph.")
    G = nx.read_gml(in_graph)

    # If a cohort is specified.
    attr = label_name.copy()
    if cohort is not None:
        attr.append('cohort')

    # Fetching labels from nodes' attributes.
    df = fetch_attributes_df(G,
                             attributes=attr)

    # Loading dataset and generating list of nodes to include.
    if distribution is not None:
        dist = pd.read_excel(distribution)
    else:
        dist = None

    # Compute weigted path metric.
    for var in label_name:
        logging.info("Computing average weighted path for variable : {}"
                     .format(var))
        avg_weighted_path, null_dist, pvalue = weightedpath(
            G,
            df,
            label_name=var,
            cohort=cohort,
            iterations=iterations,
            weight=weight,
            method=method,
            distribution=dist,
            processes=processes,
            verbose=True,
        )

        out = pd.DataFrame(
            null_dist,
            columns=[var])
        out.to_excel(f"{out_folder}/null_distributions_{var}.xlsx",
                     header=True, index=False)

        # Export metric with pvalue.
        stats = pd.DataFrame([[avg_weighted_path], [pvalue]],
                             columns=['Statistics'],
                             index=['Average weighted path', 'p-value'])
        stats.to_excel(f"{out_folder}/statistics_{var}.xlsx", header=True,
                       index=True)


if __name__ == "__main__":
    app()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging

from cyclopts import App, Parameter
import networkx as nx
import pandas as pd
from typing_extensions import Annotated

from neurostatx.io.utils import (assert_input, assert_output_dir_exist,
                                 load_df_in_any_format)
from neurostatx.network.utils import (get_nodes_and_edges,
                                      construct_attributes_dict)
from neurostatx.network.viz import (
    compute_layout,
    set_nodes_position,
    membership_distribution,
    NetworkLayout)


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def ComputeGraphNetwork(
    in_dataset: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Essential Files Options",
        ),
    ] = "./graph_results/",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    plot_distribution: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Plotting Options",
        ),
    ] = False,
    import_data: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Essential Files Options",
        ),
    ] = False,
    layout: Annotated[
        NetworkLayout,
        Parameter(
            show_default=True,
            show_choices=True,
            group="Layout Options",
        ),
    ] = NetworkLayout.Spring,
    weight: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Layout Options",
        ),
    ] = "membership",
):
    """GRAPH NETWORK CLUSTERING VISUALIZATION
    --------------------------------------
    ComputeGraphNetwork is a script computing an undirected weighted
    graph network from fuzzy clustering c-partitioned membership matrix. It is
    designed to work seemlessly with FuzzyClustering. Mapping
    membership matrices to a graph network allows the future use of graph
    theory statistics such as shortest path, betweenness centrality, etc.
    The concept of this script was initially proposed in [1].

    LAYOUT ALGORITHMS
    -----------------
    In order to generate a graph network, the nodes positions need to be
    determined in relation with their connections to other nodes (and the
    weigth of those connections). Those connections are also called edges and
    contain a weight in the case of a weighted graph network. Possible
    algorithms to choose from are :

    Kamada Kawai Layout: Use the Kamada-Kawai path-length cost-function. Not
                        the optimal solution for large network as it is
                        computer intensive. For details, see [2].

    Spectral Layout: Position is determined using the eigenvectors of the
                    graph Laplacian. For details, see [2].

    Spring Layout: Use the Fruchterman-Reingold force-directed algorithm.
                    Suitable for large network with high number of nodes.
                    For details, see [2]. This is the default method.

    IMPORTING DATA WITHIN THE .GML FILE
    -----------------------------------
    If the --import-data flag is set to True, the descriptive data will be
    imported within the .gml file. The imported data will be stored as node's
    attributes. This is useful for future use of the graph network in
    visualization scripts or in statistical analysis (view AverageWeightedPath
    or Plsr). This ensure a robust handling of data and reduce the probability
    of data mismatch between subjects.

    REFERENCES
    ----------
    [1] [Ariza-Jiménez, L., Villa, L. F., & Quintero, O. L. (2019). Memberships
        Networks for High-Dimensional Fuzzy Clustering Visualization., Applied
        Computer Sciences in Engineering (Vol. 1052, pp. 263–273). Springer
        International
        Publishing.](https://doi.org/10.1007/978-3-030-31019-6_23)

    [2] [NetworkX
    Documentation](https://networkx.org/documentation/stable/reference/drawing.html)

    EXAMPLE USAGE
    -------------
    ::

        ComputeGraphNetwork --in-dataset cluster_membership.xlsx
        --id-column subjectkey --desc-columns 1 --out-folder output/

    **For large graphs (~10 000 nodes), it might take ~5 mins to run using**
    **the spring layout and depending on your hardware.**

    Parameters
    ----------
    in_dataset : str
        Input dataset containing membership values for each clusters.
    id_column : str
        Name of the column containing the subject's ID tag. Required for
        proper handling of IDs.
    desc_columns : int
        Number of descriptive columns at the beginning of the dataset.
    out_folder : str, optional
        Path of the folder in which the results will be written. If not
        specified, current folder and default name will be used.
    verbose : bool, optional
        If true, produce verbose output.
    overwrite : bool, optional
        If true, force overwriting of existing output files.
    save_parameters : bool, optional
        If true, save the parameters used in a .txt file.
    plot_distribution : bool, optional
        If true, will plot the membership distribution and delta.
    import_data : bool, optional
        If true, will import the data from the input dataset within the graph
        network file.
    layout : NetworkLayout, optional
        Layout algorithm to determine the nodes position.
    weight : str, optional
        Name of the column containing the edge weight. Default is 'membership'.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    # Saving parameters
    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/graph_network_parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading membership matrix.
    logging.info("Loading membership data.")
    raw_df = load_df_in_any_format(in_dataset)
    descriptive_columns = [n for n in range(0, desc_columns)]

    # Creating the array.
    desc_data = raw_df[raw_df.columns[descriptive_columns]]
    clean_df = raw_df.drop(
        raw_df.columns[descriptive_columns], axis=1, inplace=False
    ).astype("float")
    df_with_ids = pd.concat([desc_data[desc_data.columns[0]], clean_df],
                            axis=1)

    # Plotting membership distributions and delta.
    if plot_distribution:
        membership_distribution(
            clean_df.values, output=f"{out_folder}/membership_distribution.png"
        )

    # Fetching dataframe of nodes and edges.
    df, _, _ = get_nodes_and_edges(df_with_ids)

    # Creating network graph.
    G = nx.from_pandas_edgelist(df, "node1", "node2", edge_attr="membership")

    # Computing graph network layout.
    logging.info("Computing graph network layout.")
    pos = compute_layout(G, layout=getattr(nx, layout), weight=weight)

    logging.info("Setting nodes position.")
    set_nodes_position(G, pos)

    if import_data:
        logging.info("Importing data within the .gml file.")
        attributes = construct_attributes_dict(desc_data,
                                               desc_data.columns[1:],
                                               id_column)
        nx.set_node_attributes(G, attributes)

    # Saving graph as a .gexf object for easy reloading.
    nx.write_gml(G, f"{out_folder}/network_graph_file.gml")


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-

from enum import Enum

from factor_analyzer import FactorAnalyzer
import matplotlib.pyplot as plt
import numpy as np
import semopy
from strenum import StrEnum


class RotationTypes(StrEnum, Enum):
    promax = "promax"
    oblimin = "oblimin"
    varimax = "varimax"
    oblimax = "oblimax"
    quartimin = "quartimin"
    quartimax = "quartimax"
    equamax = "equamax"


class MethodTypes(StrEnum, Enum):
    minres = "minres"
    ml = "ml"
    principal = "principal"


class FormattedTextPrompt(str):
    heatmap = "Evaluation of inter-variable correlation."
    screeplot = "Factor to keep are the ones with an eigenvalues over 1 (>1)."
    loadings = (
        "Plot displaying the contribution of each variable to all"
        "factors with an eigenvalues > 1."
    )
    scatterplot = (
        "Visual representation of the contribution of each variable to the"
        "first 2 factors. Scatterplot for the other factors are not produced."
    )
    semplot = (
        "Visual representation of the relationship between latent"
        "variables and indicators."
    )


def horn_parallel_analysis(
    x, output_folder, method="minres", rotation=None, nfactors=1, niter=20
):
    """
    This function is mimicking the function from the psych R package
    fa.parallel to compute the horn's parallel analysis to determine the
    appropriate number of factors to use in factorial analysis.

    Portion of this code comes from this post on stackoverflow :
    https://stackoverflow.com/questions/62303782/is-there-a-way-to-conduct-a-parallel-analysis-in-python
    and from the translation of the original function fa.parallel in the psych
    R package :
    https://github.com/cran/psych/blob/ee72f0cc2aa7c85a844e3ef63c8629096f22c35d/R/fa.parallel.R

    Results have been compared between the original R code and this function
    and no difference have been observed between the two (see pull request #11,
    https://github.com/gagnonanthony/NeuroStatX/pull/11)

    Args:
        x (np.array):               Input dataset with only variables to
                                    include in the EFA.
        method (str):               Method used to fit the model.
                                    List of possible methods:
                                        minres: Minimal Residual
                                        ml: Maximum Likelihood Factor
                                        principal: Principal Component
        rotation (str):             Rotation method to apply to the factor
                                    loadings:
                                    List of possible rotations:
                                        varimax: Orthogonal Rotation
                                        promax: Oblique Rotation
                                        oblimin: Oblique Rotation
                                        oblimax: Orthogonal Rotation
                                        quartimin: Oblique Rotation
                                        quartimax: Orthogonal Rotation
                                        equamax: Orthogonal Rotation
        nfactors (int):             Number of factors (latent variables) to
                                    extract from the data. Default is 1.
        niter (int):                Number of iterations to perform the
                                    parallel analysis. Default is 20.
    Returns:
        suggfactors (int):          Suggested number of factors to use in the
                                    factorial analysis.
        suggcomponents (int):       Suggested number of components to use in
                                    the factorial analysis.
    """

    # Getting input data dimension.
    n_sub, n_variables = x.shape

    # Initiating the factor analysis object.
    fa = FactorAnalyzer(
        n_factors=nfactors, method=method, rotation=rotation, use_smc=True
    )

    def fitting_random_data(k, n, m, sumdata):
        fa.fit(np.random.normal(size=(n, m)))
        sumdata["compeigens"] = (
            sumdata["compeigens"] + fa.get_eigenvalues()[0])
        sumdata["factoreigens"] = (
            sumdata["factoreigens"] + fa.get_eigenvalues()[1])

        return sumdata

    # Starting the iterations over random data.
    sumdata = {"compeigens": 0, "factoreigens": 0}
    for k in range(0, niter):
        sumdata = fitting_random_data(k, n_sub, n_variables, sumdata)

    sumdata["compeigens"] = sumdata["compeigens"] / niter
    sumdata["factoreigens"] = sumdata["factoreigens"] / niter

    # Fitting the real data.
    fa_values_x = np.array(fa.fit(x).get_eigenvalues())

    # Finding the optimal number of factors/components.
    suggfactors = sum((fa_values_x[1] - sumdata["factoreigens"]) > 0)
    suggcomponents = sum((fa_values_x[0] - sumdata["compeigens"]) > 0)

    # Setting up the scree plot.
    plt.figure(figsize=(10, 8))

    # Plot the eigenvalues over the number of variables.
    plt.plot([0, n_variables + 1], [1, 1], "k--", alpha=0.3)
    plt.plot(
        range(1, n_variables + 1),
        sumdata["compeigens"],
        "b",
        label="PC - random",
        alpha=0.4,
    )
    plt.scatter(range(1, n_variables + 1), fa_values_x[0], c="b", marker="o")
    plt.plot(range(1, n_variables + 1), fa_values_x[0], "b", label="PC - data")
    plt.plot(
        range(1, n_variables + 1),
        sumdata["factoreigens"],
        "g",
        label="FA - random",
        alpha=0.4,
    )
    plt.scatter(range(1, n_variables + 1), fa_values_x[1], c="g", marker="o")
    plt.plot(range(1, n_variables + 1), fa_values_x[1], "g", label="FA - data")
    plt.title("Horn's Parallel Analysis Scree Plots", {"fontsize": 16})
    plt.xlabel("Factors/Components", {"fontsize": 12})
    plt.xticks(ticks=range(1, n_variables + 1),
               labels=range(1, n_variables + 1))
    plt.ylabel("Eigenvalue", {"fontsize": 12})
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"{output_folder}/horns_parallel_screeplot.png")
    plt.close()

    return suggfactors, suggcomponents


def efa(df, method, rotation, nfactors=1):
    """
    Function to compute a simple exploratory factor analysis (EFA)
    using the factor_analyzer package.

    Args:
        df (pd.DataFrame):          Input dataset with only variables to
                                    include in the EFA.
        method (str):               Method used to fit the model.
                                    List of possible methods:
                                        minres: Minimal Residual
                                        ml: Maximum Likelihood Factor
                                        principal: Principal Component
        rotation (str):             Rotation method to apply to the factor
                                    loadings:
                                    List of possible rotations:
                                        varimax: Orthogonal Rotation
                                        promax: Oblique Rotation
                                        oblimin: Oblique Rotation
                                        oblimax: Orthogonal Rotation
                                        quartimin: Oblique Rotation
                                        quartimax: Orthogonal Rotation
                                        equamax: Orthogonal Rotation
        nfactors (int):             Number of factors (latent variables) to
                                    extract from the data. Default is 1.

    Returns:
        FactorAnalyzer:             FactorAnalyzer object containing the model.
        ev (np.array):              Eigenvalues of the model.
        v (np.array):               Eigenvalues of the model.
        scores (np.array):          Factor scores of the model.
        loadings (np.array):        Loadings of the model.
        communalities (np.array):   Communalities of the model.
    """

    # Instantiating and fitting the exploratory factorial analysis.
    efa = FactorAnalyzer(rotation=rotation, method=method,
                         n_factors=nfactors)
    efa.fit(df)

    ev, v = efa.get_eigenvalues()
    scores = efa.transform(df)
    loadings = efa.loadings_
    communalities = efa.get_communalities()

    # Return every possible information about the model and factors.
    return efa, ev, v, scores, loadings, communalities


def cfa(
    df, model
):
    """
    Used to compute a confirmatory factor analysis (CFA) to evaluate the
    goodness of fit of the model.

    This function uses the semopy package to evaluate the goodness of fit of
    the proposed model (https://semopy.com/).

    Args:
        df (pd.DataFrame):          Input dataset with only variables to
                                    include in the CFA.
        model (str):                Model description for the CFA.
    Returns:
        semopy.Model:               Model object containing the CFA.
        scores (pd.DataFrame):      Factor scores of the model.
        stats (pd.DataFrame):       Statistics of the model.
    """

    cfa = semopy.Model(model)
    cfa.fit(df)

    scores = cfa.predict_factors(df)
    stats = cfa.inspect(mode="list", what="est",
                        information="expected")

    return cfa, scores, stats



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict


download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run([
        "GraphMetrics", "-h"])

    assert ret.success


def test_eigencentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "eigencentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_closenesscentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "closenesscentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_betweennesscentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "betweennesscentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_informationcentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "informationcentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_currentflowbc(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "currentflowbc", in_graph,
        "membership", "-f"])

    assert ret.success


def test_loadcentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "loadcentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_harmoniccentrality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "harmoniccentrality", in_graph,
        "membership", "-f"])

    assert ret.success


def test_eccentricity(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "eccentricity", in_graph,
        "membership", "-f"])

    assert ret.success


def test_clustering(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "clustering", in_graph,
        "membership", "-f"])

    assert ret.success


# def test_constraint(script_runner):
#    os.chdir(os.path.expanduser(tmp_dir.name))
#    in_graph = os.path.join(get_home(), "data/graph_file.gml")

#    ret = script_runner.run([
    # "GraphMetrics",
    # "constraint", in_graph, "c1",
    # "membership", "-f"])

#    assert ret.success


# def test_effectivesize(script_runner):
#    os.chdir(os.path.expanduser(tmp_dir.name))
#    in_graph = os.path.join(get_home(), "data/graph_file.gml")

#    ret = script_runner.run([
    # "GraphMetrics",
    # "effectivesize", in_graph, "c1",
    # "membership", "-f"])

#    assert ret.success


def test_closenessvitality(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "closenessvitality", in_graph, "c1",
        "membership", "-f"])

    assert ret.success


def test_degree(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_graph = os.path.join(get_home(), "data/graph_file.gml")

    ret = script_runner.run([
        "GraphMetrics",
        "degree", in_graph,
        "membership", "-f"])

    assert ret.success



# -*- coding: utf-8 -*-

import networkx as nx
import numpy as np
import pandas as pd


def get_nodes_and_edges(df):
    """
    Function to generate a dataframe containing edges' data.

    Args:
        df (DataFrame):          Pandas DataFrame containing edges data and
                                 ids (membership
                            matrix from clustering results).

    Returns:
        Pandas DataFrame of starting node, target node and edge weights.
    """

    center_list = [f"c{i+1}" for i in range(0, len(df.columns) - 1)]
    subject_list = df[df.columns[0]]

    start_list = np.repeat(subject_list, len(df.columns) - 1)
    target_list = center_list * len(df)

    membership_data = df.drop(df.columns[0], axis=1,
                              inplace=False).astype("float")

    # Create a DataFrame Object.
    df = pd.DataFrame(
        {
            "node1": start_list,
            "node2": target_list,
            "membership": membership_data.values.flatten(),
        }
    )

    return df, subject_list, center_list


def filter_node_centroids(n):
    """
    Function to filter cluster nodes from subject's nodes.

    Args:
        n (str):        Node label.

    Returns:
        True or False
    """

    return "c" in n


def filter_node_subjects(n):
    """
    Function to filter subject nodes from cluster's nodes.

    Args:
        n (str):        Node label.

    Returns:
        True or False
    """

    return "c" not in n


def extract_subject_percentile(mat, percentile):
    """
    Function to extract subject

    Args:
        mat (Array):            Fuzzy C-partitioned membership matrix.
        percentile (float):     Percentile value.
    Return:
        label_dict:             Dictionary of binary arrays for each clusters.
    """

    # Fetching 1st and 2nd highest membership value.
    high1st = np.max(mat, axis=0)
    high2nd = np.partition(mat, -2, axis=0)[-2, :]
    delta = high1st - high2nd

    # Computing value for the Xth percentile.
    value = np.percentile(delta, percentile)

    # Labelling subjects that are over the Xth percentile value.
    label_dict = {}
    for i in range(mat.shape[0]):
        label_dict[f"c{i+1}"] = np.where(mat[i, :] > value, i + 1, 0)

    return label_dict


def construct_attributes_dict(df, labels, id_column):
    """
    Function to construct a dictionary of nodes' attributes from a DataFrame.

    Args:
        df (pd.DataFrame):      Pandas DataFrame containing nodes' attributes.
        labels (List):          List of labels to add as nodes' attributes.
        id_column (str):        Name of the column containing the subject's ID
                                tag.

    Returns:
        attributes_dict:        Dictionary of nodes' attributes.
    """

    # Set index to id_column.
    df.set_index(id_column, inplace=True)

    # Keeping only columns specified in labels.
    data_to_add = df[labels]

    # Transform to dictionary.
    attributes_dict = data_to_add.to_dict(orient="index")

    return attributes_dict


def fetch_attributes_df(G, attributes=None):
    """
    Function to fetch nodes' attributes from a graph as a DataFrame.

    Args:
        G (NetworkX Graph):     NetworkX Graph object.
        attributes (List):      List of attributes to fetch.

    Returns:
        df: pd.DataFrame        Pandas DataFrame containing nodes' attributes.
    """

    # Filter out nodes that are not subjects.
    sub_node = nx.subgraph_view(G, filter_node=filter_node_subjects)
    d = {n: G.nodes[n] for n in sub_node}

    # Filter for selected attributes.
    if len(attributes) > 0:
        d = {k: {k2: v2 for k2, v2 in v.items() if k2 in attributes}
             for k, v in d.items()}
    else:
        d = {k: {k2: v2 for k2, v2 in v.items() if k2 != 'label'}
             for k, v in d.items()}

    # Create df.
    df = pd.DataFrame.from_dict(d, orient="index")

    return df


def fetch_edge_data(G, weight='membership'):
    """
    Function to fetch edge's data from a graph as a DataFrame. This method
    works only if the graph as been created via this package.

    Args:
        G (_type_): _description_
        weight (str, optional): _description_. Defaults to 'membership'.
    """

    # Fetch the number of cluster.
    cntr_node = nx.subgraph_view(G, filter_node=filter_node_centroids)
    sub_node = nx.subgraph_view(G, filter_node=filter_node_subjects)

    # Get adjacency matrix.
    adj = np.delete(
        nx.to_numpy_array(G, weight=weight),
        [i for i in range(1, len(cntr_node) + 1)],
        axis=0)
    df = pd.DataFrame(adj[:, 1:(len(cntr_node) + 1)], index=sub_node,
                      columns=[f'Cluster {i+1}' for i in range(len(cntr_node))]
                      )

    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Import required libraries.
import coloredlogs
import dill as pickle
import logging
import sys

from cyclopts import App, Parameter
import pandas as pd
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (
    assert_input,
    assert_output_dir_exist,
    load_df_in_any_format
)
from neurostatx.utils.preprocessing import merge_dataframes
from neurostatx.statistics.utils import apply_various_models


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def ApplyModel(
    in_dataset: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    model: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    id_column: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    desc_columns: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            group="Essential Files Options",
        ),
    ] = "./ApplyModel/",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
):
    """APPLY MODEL
    -----------
    Apply a model to a dataset. Features in the dataset will be scaled prior
    to applying the model.

    CURRENTLY SUPPORTED MODELS
    --------------------------
    * SEMopy
    * FactorAnalyzer
    * sklearn

    EXAMPLE USAGE
    -------------
    ApplyModel --in-dataset dataset.xlsx --model model.pkl --id-column ID
    --desc-columns 1 --out-folder ./output -v -f -s

    Parameters
    ----------
    in_dataset : List[str]
        List of input datasets.
    model : str
        Path to the model to apply.
    id_column : str
        Column name to use as index.
    desc_columns : int
        Number of columns to use as descriptors.
    out_folder : str, optional
        Output folder.
    verbose : bool, optional
        Increase verbosity.
    save_parameters : bool, optional
        Save parameters to a file.
    overwrite : bool, optional
        Overwrite output folder if it already exists.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    logging.info(
        "Validating input files and creating output folder {}"
        .format(out_folder)
    )
    assert_input(in_dataset)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading dataset.
    logging.info("Loading {}".format(in_dataset))
    if len(in_dataset) > 1:
        if id_column is None:
            sys.exit(
                "Column name for index matching is required when inputting"
                " multiple dataframes."
            )
        dict_df = {i: load_df_in_any_format(i) for i in in_dataset}
        df = merge_dataframes(dict_df, id_column)
    else:
        df = load_df_in_any_format(in_dataset[0])

    # Assessing if NaNs are present in the dataset.
    # Disabled for now, wait until we have a solution to check only columns
    # that will be used in the model.
    # if df.isnull().values.any():
    #    raise ValueError("NaNs are present in the dataset. Please impute "
    #                     "missing values prior to applying the model.")

    descriptive_columns = [n for n in range(0, desc_columns)]
    desc_data = df.iloc[:, descriptive_columns]
    df.drop(df.columns[descriptive_columns], axis=1, inplace=True)

    # Loading model.
    logging.info("Loading model")
    with open(model, "rb") as f:
        mod = pickle.load(f)

    # Applying model.
    logging.info("Applying model")
    out = apply_various_models(df, mod)

    # Saving transformed dataset.
    logging.info("Saving transformed dataset")
    columns = desc_data.columns.append(out.columns)
    out = pd.concat([desc_data, out], axis=1, ignore_index=True)
    out.columns = columns
    out.to_excel(f"{out_folder}/transformed_dataset.xlsx", index=False)


if __name__ == "__main__":
    app()



import pandas as pd
from sklearn.impute import KNNImputer


def KNNimputation(ref_df, df, n_neighbors=5, weights='distance',
                  metric='nan_euclidean', keep_all_features=True):
    """Function to impute data in a dataset based on learned relationship
    from a reference dataset.

    This function uses the KNNImputer from the sklearn library to impute
    missing values in a dataset. The imputation is based on the relationship
    learned from a reference dataset. The reference dataset is used to
    calculate the distance between samples and the missing values are
    imputed based on the n_neighbors closest samples. Useful to complete
    data from a different population and compare both of them later on.

    ** Note: The reference dataset should not contain any missing values. **
    ** Note: The reference dataset and the dataset to impute values in should
    contain the same columns. **

    Args:
        ref_df (pd.DataFrame):              Reference dataset to learn the
                                            features' relationship from.
        df (pd.DataFrame):                  Dataset to impute.
        n_neighbors (int, optional):        Number of neighbors to use.
                                            Defaults to 5.
        weights (str, optional):            Weight function to use, possible
                                            value:
                                            * 'uniform': uniform weights. All
                                            points will have equal importance.
                                            * 'distance': Weight by the inverse
                                            of their distance.
                                            Defaults to 'distance'.
        metric (str, optional):             Distance metric for searching
                                            neighbors. Defaults to
                                            'nan_euclidean'.
        keep_all_features (bool, optional): If True, even columns containing
                                            only NaNs will be imputed.
                                            Defaults to True.

    Returns:
        pd.DataFrame:                       Imputed dataset.
    """

    # Assert columns from both datasets are named the same.
    assert ref_df.columns.all() == df.columns.all(), "Columns from both \
        datasets should be named the same. Please validate the column names."

    # Initialize the imputer.
    KNN = KNNImputer(n_neighbors=n_neighbors, metric=metric, weights=weights,
                     keep_empty_features=keep_all_features)

    # Fit the imputer to the reference dataset.
    KNN.fit(ref_df)

    # Transform data from the dataset to impute.
    out = pd.DataFrame(KNN.transform(df), columns=df.columns)

    return out


def apply_various_models(df, mod):
    """Function to apply various models to a dataset.

    Args:
        df (pd.DataFrame):              Dataframe to use.
        mod (Model):                    Model to use.

    Returns:
        y:                              Predicted values.
    """

    if "semopy" in str(type(mod)):
        y = mod.predict_factors(df)
    elif "sklearn" in str(type(mod)):
        y = pd.DataFrame(mod.transform(df),
                         columns=list(mod.get_feature_names_out()))
    elif "factor_analyzer" in str(type(mod)):
        y = pd.DataFrame(mod.transform(df),
                         columns=["factor_{}".format(i)
                                  for i in range(0, mod.n_factors)])
    else:
        raise TypeError("Model of type {} currently not supported, please"
                        "open an issue on GitHub.".format(type(mod)))

    return y



# -*- coding: utf-8 -*-

from math import sqrt
import warnings

from enum import Enum
import numpy as np
from sklearn.base import clone, is_classifier
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import (cross_val_predict, KFold, check_cv)
from sklearn.metrics import (mean_squared_error, r2_score, check_scoring)
from sklearn.utils import indexable, check_random_state
from sklearn.utils.metaestimators import _safe_split
from sklearn.utils.parallel import Parallel, delayed
from sklearn.model_selection._validation import _shuffle
from strenum import StrEnum
from tqdm import tqdm


class ScoringMethod(StrEnum, Enum):
    accuracy = "accuracy"
    balanced_accuracy = "balanced_accuracy"
    top_k_accuracy = "top_k_accuracy"
    average_precision = "average_precision"
    neg_brier_score = "neg_brier_score"
    f1 = "f1"
    f1_micro = "f1_micro"
    f1_macro = "f1_macro"
    f1_weighted = "f1_weighted"
    f1_samples = "f1_samples"
    neg_log_loss = "neg_log_loss"
    precision = "precision"
    recall = "recall"
    jaccard = "jaccard"
    roc_auc = "roc_auc"
    roc_auc_ovr = "roc_auc_ovr"
    roc_auc_ovo = "roc_auc_ovo"
    roc_auc_ovr_weighted = "roc_auc_ovr_weighted"
    roc_auc_ovo_weighted = "roc_auc_ovo_weighted"
    adjusted_rand_score = "adjusted_rand_score"
    adjusted_mutual_info_score = "adjusted_mutual_info_score"
    completeness_score = "completeness_score"
    fowlkes_mallows_score = "fowlkes_mallows_score"
    homogeneity_score = "homogeneity_score"
    mutual_info_score = "mutual_info_score"
    normalized_mutual_info_score = "normalized_mutual_info_score"
    v_measure_score = "v_measure_score"
    rand_score = "rand_score"
    explained_variance = "explained_variance"
    max_error = "max_error"
    neg_mean_absolute_error = "neg_mean_absolute_error"
    neg_mean_squared_error = "neg_mean_squared_error"
    neg_mean_squared_log_error = "neg_mean_squared_log_error"
    neg_median_absolute_error = "neg_median_absolute_error"
    neg_root_mean_squared_error = "neg_root_mean_squared_error"
    r2 = "r2"
    neg_mean_poisson_deviance = "neg_mean_poisson_deviance"
    neg_mean_gamma_deviance = "neg_mean_gamma_deviance"
    neg_mean_absolute_percentage_error = "neg_mean_absolute_percentage_error"
    d2_absolute_error_score = "d2_absolute_error_score"
    d2_pinball_score = "d2_pinball_score"
    d2_tweedie_score = "d2_tweedie_score"


class Penalty(StrEnum, Enum):
    l1 = "l1"
    l2 = "l2"
    elasticnet = "elasticnet"


class Solver(StrEnum, Enum):
    newton_cg = "newton-cg"
    newton_cholesky = "newton-cholesky"
    lbfgs = "lbfgs"
    liblinear = "liblinear"
    sag = "sag"
    saga = "saga"


def plsr_cv(X,
            Y,
            nb_comp,
            max_iter=1000,
            splits=10,
            processes=1,
            verbose=False):
    """Function to perform a PLSR model with cross-validation between a set of
    predictor and dependent variables.

    Args:
        X (pd.DataFrame):               Dataframe containing the predictor
                                        variables.
        Y (pd.DataFrame):               Dataframe containing the dependent
                                        variables.
        nb_comp (int):                  Number of components to use.
        max_iter (int, optional):       Maximum number of iterations. Defaults
                                        to 1000.
        splits (int, optional):         Number of fold to use in
                                        cross-validation. Defaults to 10.
        processes (int, optional):      Number of cpus to use during
                                        processing. Defaults to 1.
        verbose (bool, optional):       Verbose mode. Defaults to False.

    Returns:
        plsr:                           PLSR model.
        mse:                            List of mean squared errors.
        score_c:                        R2 score for the model.
        score_cv:                       R2 score for the cross-validation.
        rscore:                         Square root of the R2 score.
        mse_c:                          Mean squared error for the model.
        mse_cv:                         Mean squared error for the
                                        cross-validation.
    """
    v = True if verbose else False

    mse = []
    component = np.arange(1, nb_comp + 1)
    kf_10 = KFold(n_splits=splits, shuffle=True, random_state=1)

    # Initialize a PLSR object.
    plsr = PLSRegression(max_iter=max_iter, scale=True, tol=1e-06, copy=True)

    for i in tqdm(component, disable=v):
        plsr.n_components = i

        y_cv = cross_val_predict(plsr, X, Y, cv=kf_10, n_jobs=processes)

        mse.append(mean_squared_error(Y, y_cv))

    msemin = np.argmin(mse)

    # Fit the optimal model.
    plsr.n_components = msemin + 1
    plsr.fit(X, Y)

    y_c = plsr.predict(X)

    # Cross-validation.
    y_cv = cross_val_predict(plsr, X, Y, cv=kf_10)

    score_c = r2_score(Y, y_c)
    score_cv = r2_score(Y, y_cv)
    rscore = sqrt(score_c)
    mse_c = mean_squared_error(Y, y_c)
    mse_cv = mean_squared_error(Y, y_cv)

    return plsr, mse, score_c, score_cv, rscore, mse_c, mse_cv


def permutation_testing(estimator,
                        X,
                        Y,
                        binary=False,
                        nb_permutations=1000,
                        scoring='r2',
                        splits=10,
                        processes=1,
                        verbose=False):
    """Function to perform permutation testing on a model.

    Args:
        X (pd.DataFrame):                   Dataframe containing the predictor
                                            variables.
        Y (pd.DataFrame):                   Dataframe containing the dependent
                                            variables.
        binary (bool, optional):            If the dependent variable is
                                            binary. Defaults to False.
        nb_permutations (int, optional):    Number of iterations to perform.
                                            Defaults to 1000.
        scoring (str, optional):            Scoring method to use. Defaults
                                            to 'r2'.
        splits (int, optional):             Number of fold to use in
                                            cross-validation. Defaults to 10.
        processes (int, optional):          Number of cpus to use during
                                            processing. Defaults to 1.
        verbose (bool, optional):           Verbose mode. Defaults to False.

    Returns:
        mod:                                Model.
        score:                              Score for the model.
        coef:                               Coefficients for the model.
        perm_score:                         Scores for the permutation
                                            testing.
        score_pvalue:                       P-value for the model.
        perm_coef:                          Coefficients for the permutation
                                            testing.
        coef_pvalue:                        P-value for the coefficients.
    """
    v = 1 if verbose else 0

    # Lauching permutation testing.
    mod, score, coef, perm_score, score_pvalue, perm_coef, coef_pvalue = \
        permutation_test(
            estimator,
            X,
            Y,
            scoring=scoring,
            cv=splits,
            n_permutations=nb_permutations,
            n_jobs=processes,
            verbose=v)

    return mod, score, coef, perm_score, score_pvalue, perm_coef, coef_pvalue


def _permutation_scorer(estimator,
                        X,
                        Y,
                        groups,
                        cv,
                        scorer,
                        fit_params):
    """Core worker for permutation testing.

    Args:
        estimator (Model):              Model to use.
        X (pd.DataFrame):               Dataframe containing the predictor
                                        variables.
        Y (pd.DataFrame):               Dataframe containing the dependent
                                        variables.
        groups (pd.DataFrame):          Dataframe containing the groups.
        cv (int):                       Number of fold to use in
                                        cross-validation.
        scorer (str):                   Scoring method to use.
        fit_params (dict, optional):    Parameters to use during fitting.
                                        Defaults to None.

    Returns:
        estimator:                      Model.
        avg_score:                      Average score.
        coefficients:                   Coefficients.
    """

    fit_params = fit_params if fit_params is not None else {}
    avg_score = []
    coefficients = []

    for train, test in cv.split(X, Y, groups):
        X_train, y_train = _safe_split(estimator, X, Y, train)
        X_test, y_test = _safe_split(estimator, X, Y, test, train)
        estimator.fit(X_train, y_train, **fit_params)
        avg_score.append(scorer(estimator, X_test, y_test))
        coefficients.append(estimator.coef_.T)

    return estimator, np.mean(avg_score), np.mean(np.array(coefficients),
                                                  axis=0)


def permutation_test(estimator,
                     X,
                     Y,
                     *,
                     groups=None,
                     cv=None,
                     n_permutations=100,
                     n_jobs=None,
                     random_state=0,
                     verbose=0,
                     scoring=None,
                     fit_params=None):
    """Core function to perform permutation testing.

    Args:
        estimator (Model):                  Model to use.
        X (pd.DataFrame):                   Dataframe containing the predictor
                                            variables.
        Y (pd.DataFrame):                   Dataframe containing the dependent
                                            variables.
        groups (str, optional):             Group variable. Defaults to None.
        cv (int, optional):                 Number of fold to use in
                                            cross-validation. Defaults to None.
        n_permutations (int, optional):     Number of permutations to perform.
                                            Defaults to 100.
        n_jobs (int, optional):             Number of jobs to run in parallel.
                                            Defaults to None.
        random_state (int, optional):       Random state seed. Defaults to 0.
        verbose (int, optional):            Verbose mode. Defaults to 0.
        scoring (str, optional):            Scoring method to use. Defaults
                                            to None.
        fit_params (dict, optional):        Parameters to use during fitting.
                                            Defaults to None.

    Returns:
        mod:                                Model.
        score:                              Score for the model.
        coef:                               Coefficients for the model.
        perm_score:                         Scores for the permutation
                                            testing.
        score_pvalue:                       P-value for the model.
        perm_coef:                          Coefficients for the permutation
                                            testing.
        coef_pvalue:                        P-value for the coefficients.
    """

    warnings.filterwarnings("ignore")

    X, y, groups = indexable(X, Y, groups)

    cv = check_cv(cv, y, classifier=is_classifier(estimator))
    scorer = check_scoring(estimator, scoring=scoring)
    random_state = check_random_state(random_state)

    mod, score, coef = _permutation_scorer(clone(estimator),
                                           X, y, groups, cv, scorer,
                                           fit_params=fit_params)

    results = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(_permutation_scorer)(
            clone(estimator),
            X,
            _shuffle(y, groups, random_state),
            groups,
            cv,
            scorer,
            fit_params=fit_params,
        )
        for _ in range(n_permutations)
    )

    models, perm_score, perm_coef = zip(*results)

    perm_score = np.array(perm_score)
    perm_coef = np.array(perm_coef)
    score_pvalue = (np.sum(perm_score >= score) + 1) / (n_permutations + 1)
    coef_pvalue = (np.sum(abs(perm_coef) >= abs(coef),
                          axis=0) + 1) / (n_permutations + 1)

    return mod, score, coef, perm_score, score_pvalue, perm_coef, coef_pvalue



import unittest
import numpy as np
import pandas as pd
from factor_analyzer import FactorAnalyzer
import semopy
from sklearn.datasets import load_iris

from neurostatx.utils.factor import (horn_parallel_analysis,
                                     efa,
                                     cfa)


class TestFunctions(unittest.TestCase):

    def test_horn_parallel_analysis(self):
        # Not optimal to test this function, but it is a good start.
        # Define test data
        df = pd.DataFrame(load_iris().data, columns=load_iris().feature_names)
        output_folder = "./"
        method = "minres"
        rotation = None
        nfactors = 4
        niter = 20

        # Call function
        suggfactors, suggcomponents = horn_parallel_analysis(
            df, output_folder, method, rotation, nfactors, niter)

        # Perform assertions
        self.assertEqual(suggfactors, 3)
        self.assertEqual(suggcomponents, 1)

        # Additional assertion: check if the output file exists
        import os
        self.assertTrue(
            os.path.exists(f"{output_folder}/horns_parallel_screeplot.png"))

    def test_efa(self):
        # Not optimal to test this function, but it is a good start.
        # Define test data
        np.random.seed(123)
        df = pd.DataFrame(np.random.rand(100, 10))
        method = "minres"
        nfactors = 2
        rotation = 'varimax'

        # Call function
        efa_mod, ev, v, scores, loadings, communalities = \
            efa(df, method, rotation, nfactors)

        # Perform assertions
        self.assertIsInstance(efa_mod, FactorAnalyzer)

    def test_apply_efa_and_cfa(self):
        # Not optimal to test this function, but it is a good start.
        # Define test data
        df = pd.DataFrame(load_iris().data, columns=load_iris().feature_names)
        df.columns = ['sepal_length', 'sepal_width', 'petal_length',
                      'petal_width']
        model = ""
        model += "efa1 =~ sepal_length + petal_length\n"
        model += "efa2 =~ sepal_width + petal_length"

        # Call function
        cfa_mod, scores, stats = cfa(
            df, model)

        # Perform assertions
        self.assertIsInstance(cfa_mod, semopy.Model)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile

from neurostatx.io.download import get_home, download_data, save_files_dict

download_data(save_files_dict(), keys=["data.zip"])
tmp_dir = tempfile.TemporaryDirectory()


def test_help(script_runner):
    ret = script_runner.run(["CompareClustering", "-h"])

    assert ret.success


def test_execution_compare_clustering(script_runner):
    os.chdir(os.path.expanduser(tmp_dir.name))
    in_dataset1 = os.path.join(get_home(), "data/clusters_membership_3.xlsx")
    in_dataset2 = os.path.join(get_home(), "data/clusters_membership_3.xlsx")
    out_folder = os.path.join(get_home(), "data/ARI/")

    ret = script_runner.run([
        "CompareClustering",
        "--in-dataset", in_dataset1,
        "--in-dataset", in_dataset2,
        "--desc-columns", 1,
        "--id-column", "subjectkey",
        "--out-folder", out_folder,
        "-f"]
    )

    assert ret.success



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import coloredlogs
import logging

from cyclopts import App, Parameter
import networkx as nx
import numpy as np
from typing import List
from typing_extensions import Annotated

from neurostatx.io.utils import (assert_input, assert_output_dir_exist)
from neurostatx.network.viz import (visualize_network, create_cmap_from_list)
from neurostatx.network.utils import fetch_attributes_df


# Initializing the app.
app = App(default_parameter=Parameter(negative=()))


@app.default()
def VisualizeGraphNetwork(
    in_graph: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ],
    out_folder: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Essential Files Options",
        ),
    ] = "./visualize_network/",
    verbose: Annotated[
        bool,
        Parameter(
            "-v",
            "--verbose",
            group="Optional parameters",
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        Parameter(
            "-f",
            "--overwrite",
            group="Optional parameters",
        ),
    ] = False,
    save_parameters: Annotated[
        bool,
        Parameter(
            "-s",
            "--save_parameters",
            group="Optional parameters",
        ),
    ] = False,
    label_name: Annotated[
        List[str],
        Parameter(
            show_default=False,
            group="Label Options",
        ),
    ] = None,
    cohort: Annotated[
        int,
        Parameter(
            show_default=False,
            group="Essential Files Options",
        ),
    ] = None,
    background_alpha: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Label Options",
        ),
    ] = True,
    weight: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Network Visualization Options",
        ),
    ] = "membership",
    label_centroids: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = True,
    label_subjects: Annotated[
        bool,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = False,
    centroids_size: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 500,
    centroid_alpha: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 1,
    centroid_node_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "white",
    centroid_edge_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "black",
    subject_node_size: Annotated[
        int,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 5,
    subject_node_alpha: Annotated[
        float,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = 0.1,
    subject_node_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "darkgrey",
    subject_edge_color: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = None,
    colormap: Annotated[
        str,
        Parameter(
            show_default=True,
            group="Network Visualization Options",
        ),
    ] = "plasma",
    title: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Network Visualization Options",
        ),
    ] = "Network Graph of the clustering membership values.",
    legend_title: Annotated[
        str,
        Parameter(
            show_default=False,
            group="Network Visualization Options",
        ),
    ] = "Membership values",
):
    """VISUALIZING GRAPH NETWORK
    -------------------------
    VisualizeGraphNetwork is a command line tool to visualize the graph network
    generated from the clustering results. The script uses the NetworkX library
    to visualize the graph network. The script also provides options to label
    specific nodes based on a condition (e.g. a diagnosis, etc.). It is also
    possible to customize the graph appearance using the Network Visualization
    Options. The script is made to work hand in hand with the script
    ComputeGraphNetwork since it requires the precomputation of the nodes'
    position within the graph file.

    LABELLING GRAPH NETWORK NODES
    -----------------------------
    It is possible to label specific nodes based on a condition (e.g. a
    diagnosis, etc.). To do so, use the --label-name to indicate with binary
    node attribute to use for labelling. It is also possible to provide
    multiple label name by using --label-name x --label-name y. The script will
    output multiple graphs for each label name.

    COHORT SELECTION
    ----------------
    In some case, the graph network file can contain data from multiple cohort
    at the same time. It might be useful to single out a specific cohort to
    visualize some data. Using --cohort, the script will fetch data from
    the specified cohort and filter the --label-name. In order for the script
    to run successfully, your graph network needs to contain the cohort
    attributes. If it does not, please add it using the AddNodesAttributes
    script.

    GRAPH NETWORK CUSTOMIZATION
    ---------------------------
    To customize the graph appearance, please see the Network Visualization
    Options below. It should be noted that using subjects_labelling will crowd
    the network if it contains a high number of nodes. Also, centroids are
    labelled by default 'c1, c2, ...' and subjects 's1, s2, ...'. The script
    also exports a graph_network_file.gexf. This file can be used to further
    customize the network using other APIs such as GEPHI (see [1]).

    REFERENCES
    ----------
    [1] [GEPHI](https://gephi.org/)

    EXAMPLE USAGE
    -------------
    ::

        VisualizeGraphNetwork --in-graph graph_network.gml
        --out-folder output/ --label-name diagnosis
        --weight membership -v -f -s

    Parameters
    ----------
    in_graph : str
        Input graph network to visualize (.gml format).
    out_folder : str
        Output folder for the graph network visualization.
    verbose : bool
        Verbose mode.
    overwrite : bool
        Overwrite existing files.
    save_parameters : bool
        Save parameters to a .txt file.
    label_name : List[str]
        List of label names to subsequently use for labelling.
    cohort: int, optional
        Cohort identifier. If your graph contains data from multiple cohort,
        you can specify the cohort you want for visulization.
    background_alpha : bool
        Use background alpha for the graph.
    weight : str
        Weight to use for the graph network.
    label_centroids : bool
        Label centroids.
    label_subjects : bool
        Label subjects.
    centroids_size : int
        Size of the centroids.
    centroid_alpha : float
        Alpha of the centroids.
    centroid_node_color : str
        Color of the centroids.
    centroid_edge_color : str
        Edge color of the centroids.
    subject_node_size : int
        Size of the subjects.
    subject_node_alpha : float
        Alpha of the subjects.
    subject_node_color : str
        Color of the subjects.
    subject_edge_color : str
        Edge color of the subjects.
    colormap : str
        Colormap to use for the graph network.
    title : str
        Title of the graph network.
    legend_title : str
        Legend title of the graph network.
    """

    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        coloredlogs.install(level=logging.INFO)

    assert_input(in_graph)
    assert_output_dir_exist(overwrite, out_folder, create_dir=True)

    # Saving parameters
    if save_parameters:
        parameters = list(locals().items())
        with open(f"{out_folder}/nodes_attributes_parameters.txt", "w+") as f:
            for param in parameters:
                f.writelines(str(param))

    # Loading graph.
    logging.info("Loading graph data.")
    G = nx.read_gml(in_graph)

    # Visualizing and saving network.
    logging.info("Generating graph.")
    visualize_network(
        G,
        output=f"{out_folder}/graph_network.png",
        weight=weight,
        centroids_labelling=label_centroids,
        subjects_labelling=label_subjects,
        centroid_node_shape=centroids_size,
        centroid_alpha=centroid_alpha,
        centroid_node_color=centroid_node_color,
        centroid_edge_color=centroid_edge_color,
        subject_node_shape=subject_node_size,
        subject_alpha=subject_node_alpha,
        subject_node_color=subject_node_color,
        subject_edge_color=subject_edge_color,
        colormap=colormap,
        title=title,
        legend_title=legend_title,
    )

    # Plotting network with custom label.
    if label_name is not None:

        logging.info("Constructing graph(s) with custom labels.")

        if cohort is not None:
            label_name.append('cohort')

        # Loading df.
        df = fetch_attributes_df(G, attributes=label_name)

        # Fetching data for label as array.
        for label in label_name:
            if label == 'cohort':
                continue

            # If subject is not within the specified cohort, impute 0.
            if cohort is not None:
                df.loc[:, label] = np.where(df.loc[:, label] == 1,
                                            np.where(
                                                df.loc[:, 'cohort'] == cohort,
                                                1,
                                                0),
                                            0)
            labels = df[label]

            nodes_cmap = create_cmap_from_list(labels)

            if background_alpha:
                sub_alpha = []
                for i in nodes_cmap:
                    if isinstance(i, str):
                        sub_alpha.append(0.1)
                    else:
                        sub_alpha.append(1)
            else:
                sub_alpha = np.array([1] * len(nodes_cmap))

            visualize_network(
                G,
                output=f"{out_folder}/graph_network_{label}.png",
                weight=weight,
                centroids_labelling=label_centroids,
                subjects_labelling=label_subjects,
                centroid_node_shape=centroids_size,
                centroid_alpha=centroid_alpha,
                centroid_node_color=centroid_node_color,
                centroid_edge_color=centroid_edge_color,
                subject_node_shape=subject_node_size,
                subject_alpha=sub_alpha,
                subject_node_color=nodes_cmap,
                subject_edge_color=subject_edge_color,
                colormap=colormap,
                title=f"{title} with {label} subjects colored.",
                legend_title=legend_title,
            )


if __name__ == "__main__":
    app()



import unittest

import numpy as np
import pandas as pd
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score

from neurostatx.clustering.metrics import (compute_evaluation_metrics,
                                           compute_knee_location,
                                           compute_sse,
                                           compute_gap_stats,
                                           find_optimal_gap,
                                           compute_rand_index)


class TestComputeEvaluationMetrics(unittest.TestCase):

    def setUp(self):
        # Generate synthetic data with known characteristics
        self.X, true_labels = make_blobs(n_samples=300, centers=3,
                                         cluster_std=0.5,
                                         random_state=42)

        # Assume known silhouette score, Calinski-Harabasz index, and
        # Davies-Bouldin index values
        self.expected_ss = 0.92
        self.expected_chi = 20500.0
        self.expected_dbi = 0.10

        # Assuming labels are known in this synthetic data
        self.labels = true_labels

    def test_compute_evaluation_metrics(self):
        # Test the function with synthetic data and known characteristics
        ss, chi, dbi = compute_evaluation_metrics(self.X, self.labels)

        # Check if the returned values are close to the expected values
        self.assertAlmostEqual(ss, self.expected_ss, places=2)
        self.assertAlmostEqual(chi, self.expected_chi, places=-3)
        self.assertAlmostEqual(dbi, self.expected_dbi, places=1)


class TestComputeKneeLocation(unittest.TestCase):

    def setUp(self):
        # Elbow data.
        elbow_point = 5
        self.values = np.zeros(10)

        # Generate data before the elbow point with a linear relationship
        self.values[:elbow_point] = -2 * np.arange(1, elbow_point + 1)

        # Generate data after the elbow point with a different slope
        self.values[elbow_point:] = -0.5 * np.arange(1, 10 - elbow_point + 1)

        # Add random noise to the data
        self.values += np.random.randn(10) * 0.1

    def test_compute_knee_location(self):
        # Check if the returned elbow location is close to the expected value
        elbow = compute_knee_location(self.values)
        self.assertEqual(elbow - 1, 5)


class TestComputeSSE(unittest.TestCase):

    def setUp(self):
        # Generate random centroids
        self.centroids = np.random.rand(3, 2)

        # Initialize variables
        iteration = 0
        scale_factor = 1.0
        self.target_wss = 100.0

        self.data = np.zeros((1000, 2))
        self.cluster_assignments = np.random.randint(0, 3, size=1000)
        # Repeat until the difference between actual and target WSS falls
        # within the tolerance or max iterations are reached
        while True:
            # Generate data points around centroids
            for i in range(1000):
                centroid_index = self.cluster_assignments[i]
                centroid = self.centroids[centroid_index]
                # Adjust covariance matrix for tighter clusters
                covariance_matrix = np.eye(2) * 0.1
                self.data[i] = np.random.multivariate_normal(centroid,
                                                             covariance_matrix)

            # Compute actual WSS
            self.actual_wss = np.sum(
                np.linalg.norm(
                    self.data - self.centroids[self.cluster_assignments],
                    axis=1) ** 2)

            # Check if the actual WSS falls within the desired range
            if abs(self.actual_wss - self.target_wss) \
                    <= 10.0 or iteration >= 500:
                break

            # Adjust the scale factor based on the difference between actual
            # and target WSS
            scale_factor *= np.sqrt(self.target_wss / self.actual_wss)
            self.data *= scale_factor

            # Increment iteration counter
            iteration += 1

    def test_compute_see(self):
        # Check if the returned WSS is close to the expected value.
        wss = compute_sse(self.data, self.centroids, self.cluster_assignments)
        self.assertAlmostEqual(wss, self.actual_wss, places=2)


class TestComputeGapStats(unittest.TestCase):

    def setUp(self):
        # Simulate gap values.
        self.gap_values = [0.1, 0.2, 0.3, 0.2, 0.1]
        self.gap_std_devs = [0.01, 0.02, 0.03, 0.02, 0.01]

        # Simulate data for gap statistic.
        np.random.seed(0)
        self.X = np.random.rand(1000, 2)
        self.wss = 100.0

    def test_find_optimal_gap(self):
        # Test the function with simulated data
        optimal_k = find_optimal_gap(self.gap_values, self.gap_std_devs)
        self.assertEqual(optimal_k, 2)

    def test_compute_gap_stats(self):
        # Test the function with simulated data
        np.random.seed(0)
        gap, _ = compute_gap_stats(self.X, self.wss, 1000, 3)

        # Assert that the computed gap statistic is equal or almost equal to
        # the expected value
        self.assertGreaterEqual(gap, -0.45)
        self.assertLessEqual(gap, -0.40)


class TestComputeRandIndex(unittest.TestCase):

    def setUp(self):
        # Generate synthetic data with known clustering assignments
        num_samples = 100
        num_clusters = 3
        self.clustering_dict = {}
        for i in range(num_clusters):
            self.clustering_dict[f'Cluster_{i}'] = \
                pd.DataFrame(np.random.rand(num_samples, num_clusters))

        # Calculate the expected Rand Index
        self.expected_ari = np.zeros((num_clusters, num_clusters))
        for i, (_, arr1) in enumerate(self.clustering_dict.items()):
            for j, (_, arr2) in enumerate(self.clustering_dict.items()):
                labels1 = np.argmax(arr1, axis=1)
                labels2 = np.argmax(arr2, axis=1)
                self.expected_ari[i, j] = adjusted_rand_score(labels1, labels2)

    def test_compute_rand_index(self):
        # Test the function with synthetic data
        ari = compute_rand_index(self.clustering_dict)

        # Compare the computed Rand Index with the expected value
        np.testing.assert_array_almost_equal(ari, self.expected_ari)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import numpy as np

from neurostatx.statistics.harmonization import neuroCombat


class TestHarmonization(unittest.TestCase):

    def test_neuroCombat_noneb(self):
        # Define a dataset to harmonize.
        df = pd.DataFrame({'A': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0],
                           'B': [5.0, 4.0, 3.0, 3.0, 2.0, 1.0],
                           'C': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0]})

        # Define a dataset with batch information.
        covars = pd.DataFrame([1, 1, 1, 2, 2, 2], columns=['batch'])

        # Harmonize the dataset.
        out = neuroCombat(df.values.T,
                          covars=covars,
                          batch_col='batch',
                          eb=False,
                          ref_batch=1)

        expected_output = pd.DataFrame({'A': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                        'B': [5.0, 4.0, 3.0, 5.0, 4.0, 3.0],
                                        'C': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0]})

        # Compare the output with the expected output.
        np.testing.assert_allclose(out['data'],
                                   expected_output.values.T,
                                   atol=0.2)

    def test_neuroCombat_nonparam(self):
        # Define a dataset to harmonize.
        df = pd.DataFrame({'A': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0],
                           'B': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0],
                           'C': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0]})

        # Define a dataset with batch information.
        covars = pd.DataFrame([1, 1, 1, 2, 2, 2], columns=['batch'])

        # Harmonize the dataset.
        out = neuroCombat(df.values.T,
                          covars=covars,
                          batch_col='batch',
                          eb=True,
                          parametric=False,
                          ref_batch=1)

        expected_output = pd.DataFrame({'A': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                        'B': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                        'C': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0]})

        # Compare the output with the expected output.
        np.testing.assert_allclose(out['data'],
                                   expected_output.values.T,
                                   atol=0.2)

    def test_neuroCombat_param(self):
        # Define a dataset to harmonize.
        df = pd.DataFrame({'A': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0],
                           'B': [5.0, 4.0, 3.0, 3.0, 2.0, 1.0],
                           'C': [1.0, 2.0, 3.0, 3.0, 4.0, 5.0]})

        # Define a dataset with batch information.
        covars = pd.DataFrame([1, 1, 1, 2, 2, 2], columns=['batch'])

        # Harmonize the dataset.
        out = neuroCombat(df.values.T,
                          covars=covars,
                          batch_col='batch',
                          eb=True,
                          ref_batch=1)

        expected_output = pd.DataFrame({'A': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                                        'B': [5.0, 4.0, 3.0, 5.0, 4.0, 3.0],
                                        'C': [1.0, 2.0, 3.0, 1.0, 2.0, 3.0]})

        # Compare the output with the expected output.
        np.testing.assert_allclose(out['data'],
                                   expected_output.values.T,
                                   atol=0.35)



# -*- coding: utf-8 -*-

from kneed import KneeLocator
import numpy as np
from sklearn.metrics import (
    silhouette_score,
    calinski_harabasz_score,
    davies_bouldin_score,
    adjusted_rand_score
)
import skfuzzy as fuzz


def compute_evaluation_metrics(X, labels, metric="euclidean"):
    """
    Function to compute a variety of metrics to evaluate the goodness of fit
    of a clustering model.

    Args:
        X (Array-like):             Data from clustering algorithm to
                                    derive metrics from.
        labels (List):              List of labels.
        metric (str, optional):     Distance metric to use. Defaults
                                    to 'euclidean'. Accept options from
                                    sklearn.metrics.pairwise.pairwise_distances.

    Returns:
        ss:                         Silhouette Score (SS).
        chi:                        Calinski Harabasz Score (CHI).
        dbi:                        Davies Bouldin Score (DBI).
    """

    # Storing Silhouette score.
    ss = silhouette_score(X, labels, metric=metric)

    # Storing Calinski-Harabasz Indices (CHI).
    chi = calinski_harabasz_score(X, labels)

    # Storing the Davies-Bouldin Index (DBI).
    dbi = davies_bouldin_score(X, labels)

    return ss, chi, dbi


def compute_knee_location(lst, direction="decreasing"):
    """
    Funtion to compute the Elbow location using the Kneed package.

    Args:
        lst (List):     List of values representing the indicators to identify
                        the elbow location.

    Returns:
        elbow:          Elbow location.
    """

    knee = KneeLocator(
        range(2, len(lst) + 2), lst, S=1, curve="convex", direction=direction
    )
    elbow = knee.elbow

    return elbow


def compute_sse(X, cntr, labels):
    """
    Function to compute within cluster sum of square error (WSS).
    Adapted from :
    https://towardsdatascience.com/how-to-determine-the-right-number-of-clusters-with-code-d58de36368b1

    Args:
        X (Array):          Original data (S, N)
        cntr (Array):       Centroid points (N, F)
        labels (Array):     Discrete labels (S,)

    Returns:
        WSS:                Within Sum-of-Squares Error (WSS)
    """

    WSS = 0
    for k in np.unique(labels):
        data_k = X[labels == k]
        WSS += np.sum([np.abs(data_k - cntr[k]) ** 2])

    return WSS


def compute_gap_stats(
    X,
    wss,
    nrefs,
    n_cluster,
    m=2,
    error=1e-6,
    maxiter=1000,
    metric="euclidean",
    init=None,
):
    """
    Function to compute the GAP Statistics to determine the optimal number of
    clusters.
    Adapted from :
    https://towardsdatascience.com/cheat-sheet-to-implementing-7-methods-for-selecting-optimal-number-of-clusters-in-python-898241e1d6ad
    and https://github.com/milesgranger/gap_statistic

    Args:
        X (array):                      Data array on which clustering will be
                                        computed.
        wss (float):                    Within Cluster Sum of Squared Error
                                        (WSS) for this clustering model.
        nrefs (int):                    Number of random reference data to
                                        generate and average.
        n_cluster (int):                Number of cluster in for this model.
        m (int, optional):              Exponentiation value as used in the
                                        main script. Defaults to 2.
        error (float, optional):        Convergence error threshold. Defaults
                                        to 1E-6.
        maxiter (int, optional):        Maximum iterations to perform.
                                        Defaults to 1000.
        metric (str, optional):         Distance metric to use. Defaults to
                                        'euclidean'.
        init (array, optional):         Initial fuzzy c-partitioned matrix.
                                        Defaults to None.

    Returns:
        gap:                            GAP Statistics.
        sk:                             Standard deviation of the GAP
                                        statistic.
    """

    refDisps = np.zeros(nrefs)
    a, b = X.min(axis=0, keepdims=True), X.max(axis=0, keepdims=True)

    for i in range(nrefs):
        randomRef = np.random.random_sample(size=X.shape) * (b - a) + a

        cntr, u, u0, d, jm, p, fpc = fuzz.cmeans(
            randomRef.T,
            n_cluster,
            m=m,
            error=error,
            maxiter=maxiter,
            metric=metric,
            init=init,
        )

        labels = np.argmax(u, axis=0)
        refDisps[i] = compute_sse(randomRef, cntr, labels)

    gap = np.mean(np.log(refDisps)) - np.log(wss)

    # Compute standard deviation.
    sdk = np.sqrt(np.mean((np.log(refDisps) - np.mean(np.log(refDisps))
                           ) ** 2.0))
    sk = np.sqrt(1.0 + 1.0 / nrefs) * sdk

    return gap, sk


def find_optimal_gap(gap, sk):
    """
    Function to find the optimal k number based on the GAP statistics using
    the method from Tibshirani R. et al., 2001
    (https://hastie.su.domains/Papers/gap.pdf). Highlights the first
    k value where GAP[k] >= GAP[k+1] - SD[k+1].

    Args:
        gap (ndarray):              Ndarray of GAP statistics values for a
                                    range of k clusters.
        sk (ndarray):               Ndarray of standard deviation for each GAP
                                    values.

    Returns:
        optimal:                    Optimal k value.
    """

    for i in range(len(gap)):
        if i + 1 in range(len(gap)):
            if gap[i] >= (gap[i + 1] - sk[i + 1]):
                optimal = i
                break
        else:
            optimal = i
            break

    return optimal


def compute_rand_index(dict):
    """
    Compute the adjusted Rand Index from a list of fuzzy membership matrices
    using sklearn.metrics.adjusted_rand_score. A defuzzification step is
    required since this method applies only to crisp clusters.

    Args:
        dict (dictionnary):             Dictonnary containing all dataframes.

    Return:
        Ndarray                         Symmetric ndarray.
    """

    ari = []
    keys = list(dict.keys())
    for k in keys:
        for k2 in keys:
            val1 = dict[k].values.argmax(axis=1)
            val2 = dict[k2].values.argmax(axis=1)
            ari.append(adjusted_rand_score(val1, val2))

    return np.array(ari).reshape(len(keys), len(keys))



# -*- coding: utf-8 -*-
import os
import logging
import zipfile
import pathlib

import gdown


def save_files_dict():
    """Getting dictionary file list from GDrive."""
    return {"data.zip": "1lMHNX_y5n6phqzKNUKbVIvd_onv-zxkk"}


def download_file_from_GDrive(id, destination):
    """
    Function to download a zip file from a Google Drive.
    :param id:
    :param destination:
    :return:
    """
    gdown.download(id=id, output=destination, quiet=True)


def get_home():
    """Set home directory of NeuroStatX for testing."""
    if "NeuroStatX_HOME" in os.environ:
        neurostatx_home = os.environ["NeuroStatX_HOME"]
    else:
        neurostatx_home = os.path.join(os.path.expanduser("~"), ".neurostatx")
    return neurostatx_home


def download_data(files_dict, keys=None):
    """
    Function to download and extract data with gdown.
    Using structure from:
    https://github.com/scilus/scilpy/blob/master/scilpy/io/fetcher.py
    :param files_dict:
    :param keys:
    :return:
    """

    neurostatx_home = get_home()

    if not os.path.exists(neurostatx_home):
        os.makedirs(neurostatx_home)

    if keys is None:
        keys = files_dict.keys()
    elif isinstance(keys, str):
        keys = [keys]
    for f in keys:
        key = files_dict[f]
        full_path = os.path.join(neurostatx_home, f)
        full_path_no_ext, ext = os.path.splitext(full_path)

        if not os.path.isdir(full_path_no_ext):
            if ext == ".zip" and not os.path.isdir(full_path_no_ext):
                logging.warning(
                    "Downloading file and extracting {} from id {} to {}"
                    .format(
                        f, key, neurostatx_home
                    )
                )

                download_file_from_GDrive(key, full_path)

                try:
                    z = zipfile.ZipFile(full_path)
                    zipinfos = z.infolist()
                    root_dir = (pathlib.Path(
                                zipinfos[0].filename).parts[0] + "/")
                    assert all([s.startswith(root_dir) for s in z.namelist()])
                    with zipfile.ZipFile(full_path, "r") as zip_ref:
                        zip_ref.extractall(neurostatx_home)

                except AssertionError:
                    z.extractall(full_path)
            else:
                raise NotImplementedError("Function was expected a zip file.")

        else:
            logging.warning("Data already downloaded.")
